from __future__ import annotations

import sys

import click
import humanize
from tabulate import tabulate

from ai.backend.cli.types import ExitCode
from ai.backend.client.func.vfolder import _default_list_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..pretty import print_error
from ..types import CLIContext
from ..vfolder import vfolder as user_vfolder
from . import admin


@admin.group()
def vfolder() -> None:
    """
    VFolder administration commands.
    """


def _list_cmd(docs: str = None):
    @pass_ctx_obj
    @click.option("-g", "--group", type=str, default=None, help="Filter by group ID.")
    @click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
    @click.option("--order", default=None, help="Set the query ordering expression.")
    @click.option("--offset", default=0, help="The index of the current page start for pagination.")
    @click.option("--limit", type=int, default=None, help="The page size for pagination.")
    def list(ctx: CLIContext, group, filter_, order, offset, limit) -> None:
        """
        List virtual folders.
        """
        try:
            with Session() as session:
                fetch_func = lambda pg_offset, pg_size: session.VFolder.paginated_list(
                    group,
                    fields=_default_list_fields,
                    page_offset=pg_offset,
                    page_size=pg_size,
                    filter=filter_,
                    order=order,
                )
                ctx.output.print_paginated_list(
                    fetch_func,
                    initial_page_offset=offset,
                    page_size=limit,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)

    if docs is not None:
        list.__doc__ = docs
    return list


user_vfolder.command()(_list_cmd())
vfolder.command()(_list_cmd())


@vfolder.command()
def list_hosts():
    """
    List all mounted hosts from virtual folder root.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            resp = session.VFolder.list_all_hosts()
            print("Default vfolder host: {}".format(resp["default"]))
            print("Mounted hosts: {}".format(", ".join(resp["allowed"])))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("vfolder_host")
def perf_metric(vfolder_host):
    """
    Show the performance statistics of a vfolder host.
    (superadmin privilege required)

    A vfolder host consists of a string of the storage proxy name and the volume name
    separated by a colon. (e.g., "local:volume1")
    """
    with Session() as session:
        try:
            resp = session.VFolder.get_performance_metric(vfolder_host)
            print(
                tabulate(
                    [
                        (k, humanize.naturalsize(v, binary=True) if "bytes" in k else f"{v:.2f}")
                        for k, v in resp["metric"].items()
                    ],
                    headers=("Key", "Value"),
                )
            )
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.option(
    "-a", "--agent-id", type=str, default=None, help="Target agent to fetch fstab contents."
)
def get_fstab_contents(agent_id):
    """
    Get contents of fstab file from a node.
    (superadmin privilege required)

    If agent-id is not specified, manager's fstab contents will be returned.
    """
    with Session() as session:
        try:
            resp = session.VFolder.get_fstab_contents(agent_id)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        print(resp)


@vfolder.command()
def list_mounts():
    """
    List all mounted hosts in virtual folder root.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            resp = session.VFolder.list_mounts()
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        print("manager")
        for k, v in resp["manager"].items():
            print(" ", k, ":", v)
        print("\nagents")
        for aid, data in resp["agents"].items():
            print(" ", aid)
            for k, v in data.items():
                print("   ", k, ":", v)


@vfolder.command()
@click.argument("fs-location", type=str)
@click.argument("name", type=str)
@click.option("-o", "--options", type=str, default=None, help="Mount options.")
@click.option("--edit-fstab", is_flag=True, help="Edit fstab file to mount permanently.")
def mount_host(fs_location, name, options, edit_fstab):
    """
    Mount a host in virtual folder root.
    (superadmin privilege required)

    \b
    FS-LOCATION: Location of file system to be mounted.
    NAME: Name of mounted host.
    """
    with Session() as session:
        try:
            resp = session.VFolder.mount_host(name, fs_location, options, edit_fstab)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        print("manager")
        for k, v in resp["manager"].items():
            print(" ", k, ":", v)
        print("agents")
        for aid, data in resp["agents"].items():
            print(" ", aid)
            for k, v in data.items():
                print("   ", k, ":", v)


@vfolder.command()
@click.argument("name", type=str)
@click.option("--edit-fstab", is_flag=True, help="Edit fstab file to mount permanently.")
def umount_host(name, edit_fstab):
    """
    Unmount a host from virtual folder root.
    (superadmin privilege required)

    \b
    NAME: Name of mounted host.
    """
    with Session() as session:
        try:
            resp = session.VFolder.umount_host(name, edit_fstab)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        print("manager")
        for k, v in resp["manager"].items():
            print(" ", k, ":", v)
        print("agents")
        for aid, data in resp["agents"].items():
            print(" ", aid)
            for k, v in data.items():
                print("   ", k, ":", v)


@vfolder.command
def list_shared_vfolders():
    """
    List all shared vfolder.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            resp = session.VFolder.list_shared_vfolders()
            result = resp.get("shared", [])
            for _result in result:
                print(
                    'Virtual folder "{0}" (ID: {1})'.format(
                        _result["vfolder_name"], _result["vfolder_id"]
                    )
                )
                print("- Owner: {0}".format(_result["owner"]))
                print("- Status: {0}".format(_result["status"]))
                print("- Permission: {0}".format(_result["perm"]))
                print("- Folder Type: {0}".format(_result["type"]))
                shared_to = _result.get("shared_to", {})
                if shared_to:
                    print("- Shared to:")
                    for k, v in shared_to.items():
                        print("\t- {0}: {1}\n".format(k, v))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command
@click.argument("vfolder_id", type=str)
def shared_vfolder_info(vfolder_id):
    """Show the vfolder permission information of the given virtual folder.

    \b
    VFOLDER_ID: ID of a virtual folder.
    """
    with Session() as session:
        try:
            resp = session.VFolder.shared_vfolder_info(vfolder_id)
            result = resp.get("shared", [])
            if result:
                _result = result[0]
                print(
                    'Virtual folder "{0}" (ID: {1})'.format(
                        _result["vfolder_name"], _result["vfolder_id"]
                    )
                )
                print("- Owner: {0}".format(_result["owner"]))
                print("- Status: {0}".format(_result["status"]))
                print("- Permission: {0}".format(_result["perm"]))
                print("- Folder Type: {0}".format(_result["type"]))
                shared_to = _result.get("shared_to", {})
                if shared_to:
                    print("- Shared to:")
                    for k, v in shared_to.items():
                        print("\t- {0}: {1}\n".format(k, v))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("vfolder_id", type=str)
@click.argument("user_id", type=str)
@click.option(
    "-p", "--permission", type=str, metavar="PERMISSION", help="Folder's innate permission."
)
def update_shared_vf_permission(vfolder_id, user_id, permission):
    """
    Update permission for shared vfolders.

    \b
    VFOLDER_ID: ID of a virtual folder.
    USER_ID: ID of user who have been granted access to shared vFolder.
    PERMISSION: Permission to update. "ro" (read-only) / "rw" (read-write) / "wd" (write-delete).
    """
    with Session() as session:
        try:
            resp = session.VFolder.update_shared_vfolder(vfolder_id, user_id, permission)
            print("Updated.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        print(resp)


@vfolder.command()
@click.argument("vfolder_id", type=str)
@click.argument("user_id", type=str)
def remove_shared_vf_permission(vfolder_id, user_id):
    """
    Remove permission for shared vfolders.

    \b
    VFOLDER_ID: ID of a virtual folder.
    USER_ID: ID of user who have been granted access to shared vFolder.
    """
    with Session() as session:
        try:
            resp = session.VFolder.update_shared_vfolder(vfolder_id, user_id, None)
            print("Removed.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)
        print(resp)



from __future__ import annotations

import decimal
import json
import textwrap
from typing import Any, Mapping, Optional

import humanize

from .types import AbstractOutputFormatter, FieldSpec


def format_stats(raw_stats: Optional[str], indent="") -> str:
    if raw_stats is None:
        return "(unavailable)"
    stats = json.loads(raw_stats)
    text = "\n".join(f"- {k + ': ':18s}{v}" for k, v in stats.items())
    return "\n" + textwrap.indent(text, indent)


def format_multiline(value: Any, indent_length: int) -> str:
    buf = []
    for idx, line in enumerate(str(value).strip().splitlines()):
        if idx == 0:
            buf.append(line)
        else:
            buf.append((" " * indent_length) + line)
    return "\n".join(buf)


def format_nested_dicts(value: Mapping[str, Mapping[str, Any]]) -> str:
    """
    Format a mapping from string keys to sub-mappings.
    """
    rows = []
    if not value:
        rows.append("(empty)")
    else:
        for outer_key, outer_value in value.items():
            if isinstance(outer_value, dict):
                if outer_value:
                    rows.append(f"+ {outer_key}")
                    inner_rows = format_nested_dicts(outer_value)
                    rows.append(textwrap.indent(inner_rows, prefix="  "))
                else:
                    rows.append(f"+ {outer_key}: (empty)")
            else:
                if outer_value is None:
                    rows.append(f"- {outer_key}: (null)")
                else:
                    rows.append(f"- {outer_key}: {outer_value}")
    return "\n".join(rows)


def format_value(value: Any) -> str:
    if value is None:
        return "(null)"
    if isinstance(value, (dict, list, set)) and not value:
        return "(empty)"
    return str(value)


class OutputFormatter(AbstractOutputFormatter):
    """
    The base implementation of output formats.
    """

    def format_console(self, value: Any, field: FieldSpec) -> str:
        if value is None:
            return "(null)"
        if isinstance(value, (dict, list, set)) and not value:
            return "(empty)"
        elif isinstance(value, dict):
            return (
                "{"
                + ", ".join(f"{k}: {self.format_console(v, field)}" for k, v in value.items())
                + "}"
            )
        elif isinstance(value, (list, tuple, set)):
            return "[" + ", ".join(self.format_console(v, field) for v in value) + "]"
        return str(value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        if value is None:
            return None
        if isinstance(value, decimal.Decimal):
            return str(value)
        elif isinstance(value, dict):
            return {k: self.format_json(v, field) for k, v in value.items()}
        elif isinstance(value, (list, tuple)):
            return [self.format_json(v, field) for v in value]
        return value


class NestedDictOutputFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        if value is None:
            return "(null)"
        value = json.loads(value)
        return format_nested_dicts(value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        if value is None:
            return None
        return json.loads(value)


class MiBytesOutputFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        value = round(value / 2**20, 1)
        return super().format_console(value, field)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        value = round(value / 2**20, 1)
        return super().format_json(value, field)


class SizeBytesOutputFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        value = humanize.naturalsize(value, binary=True)
        return super().format_console(value, field)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        value = humanize.naturalsize(value, binary=True)
        return super().format_json(value, field)


class SubFieldOutputFormatter(OutputFormatter):
    def __init__(self, subfield_name: str) -> None:
        self._subfield_name = subfield_name

    def format_console(self, value: Any, field: FieldSpec) -> str:
        return super().format_console(value[self._subfield_name], field)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return super().format_json(value[self._subfield_name], field)


class ResourceSlotFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        value = json.loads(value)
        if mem := value.get("mem"):
            value["mem"] = humanize.naturalsize(mem, binary=True, gnu=True)
        return ", ".join(f"{k}:{v}" for k, v in value.items())

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return json.loads(value)


default_output_formatter = OutputFormatter()
nested_dict_formatter = NestedDictOutputFormatter()
mibytes_output_formatter = MiBytesOutputFormatter()
resource_slot_formatter = ResourceSlotFormatter()
sizebytes_output_formatter = SizeBytesOutputFormatter()


class AgentStatFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        try:
            raw_stats = json.loads(value)
        except TypeError:
            return ""

        value_formatters = {
            "bytes": lambda m: "{} / {}".format(
                humanize.naturalsize(int(m["current"]), binary=True),
                humanize.naturalsize(int(m["capacity"]), binary=True),
            ),
            "Celsius": lambda m: "{:,} C".format(
                float(m["current"]),
            ),
            "bps": lambda m: "{}/s".format(
                humanize.naturalsize(float(m["current"])),
            ),
            "pct": lambda m: "{} %".format(
                m["pct"],
            ),
        }

        def format_value(metric):
            formatter = value_formatters.get(
                metric["unit_hint"],
                lambda m: "{} / {} {}".format(
                    m["current"],
                    m["capacity"],
                    m["unit_hint"],
                ),
            )
            return formatter(metric)

        bufs = []
        node_metric_bufs = []
        for stat_key, metric in raw_stats["node"].items():
            if stat_key == "cpu_util":
                num_cores = len(raw_stats["devices"]["cpu_util"])
                if metric["pct"] is None:
                    node_metric_bufs.append(f"{stat_key}: (calculating...) % ({num_cores} cores)")
                else:
                    node_metric_bufs.append(f"{stat_key}: {metric['pct']} % ({num_cores} cores)")
            else:
                node_metric_bufs.append(f"{stat_key}: {format_value(metric)}")
        bufs.append(", ".join(node_metric_bufs))
        dev_metric_bufs = []
        for stat_key, per_dev_metric in raw_stats["devices"].items():
            dev_metric_bufs.append(f"+ {stat_key}")
            if stat_key == "cpu_util" and len(per_dev_metric) > 8:
                dev_metric_bufs.append(
                    "  - (per-core stats hidden for large CPUs with more than 8 cores)",
                )
            else:
                for dev_id, metric in per_dev_metric.items():
                    dev_metric_bufs.append(
                        f"  - {dev_id}: {format_value(metric)}",
                    )
        bufs.append("\n".join(dev_metric_bufs))
        return "\n".join(bufs)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        # TODO: improve
        return self.format_console(value, field)


class GroupListFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        return ", ".join(g["name"] for g in value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return value


class KernelStatFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        return format_stats(value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return value


class NestedObjectFormatter(OutputFormatter):
    def format_json(self, value: Any, field: FieldSpec) -> Any:
        assert isinstance(value, list)
        return [
            {
                f.alt_name: f.formatter.format_json(item[f.field_name], f)
                for f in field.subfields.values()
            }
            for item in value
        ]


def _fit_multiline_in_cell(text: str, indent: str) -> str:
    if "\n" in text:
        return "\n" + textwrap.indent(text, indent)
    else:
        return text


class ContainerListFormatter(NestedObjectFormatter):
    def format_console(self, value: Any, field: FieldSpec, indent="") -> str:
        assert isinstance(value, list)
        if len(value) == 0:
            text = "(no sub-containers belonging to the session)"
        else:
            text = ""
            for item in value:
                text += f"+ {item['id']}\n"
                text += "\n".join(
                    f"  - {f.humanized_name}: "
                    f"{_fit_multiline_in_cell(f.formatter.format_console(item[f.field_name], f), '    ')}"  # noqa
                    for f in field.subfields.values()
                    if f.field_name != "id"
                )
        return textwrap.indent(text, indent)


class DependencyListFormatter(NestedObjectFormatter):
    def format_console(self, value: Any, field: FieldSpec, indent="") -> str:
        assert isinstance(value, list)
        if len(value) == 0:
            text = "(no dependency tasks)"
        else:
            text = ""
            for item in value:
                text += f"+ {item['name']} ({item['id']})\n"
                text += "\n".join(
                    f"  - {f.humanized_name}: "
                    f"{_fit_multiline_in_cell(f.formatter.format_console(item[f.field_name], f), '    ')}"  # noqa
                    for f in field.subfields.values()
                    if f.field_name not in ("id", "name")
                )
        return textwrap.indent(text, indent)



from typing import Iterable, Optional, Sequence

from ai.backend.client.output.fields import keypair_resource_policy_fields
from ai.backend.client.output.types import FieldSpec

from ..session import api_session
from .base import BaseFunction, api_function, resolve_fields

__all__ = "KeypairResourcePolicy"

_default_list_fields = (
    keypair_resource_policy_fields["name"],
    keypair_resource_policy_fields["created_at"],
    keypair_resource_policy_fields["total_resource_slots"],
    keypair_resource_policy_fields["max_concurrent_sessions"],
    keypair_resource_policy_fields["max_vfolder_count"],
    keypair_resource_policy_fields["max_vfolder_size"],
    keypair_resource_policy_fields["idle_timeout"],
    keypair_resource_policy_fields["max_containers_per_session"],
    keypair_resource_policy_fields["allowed_vfolder_hosts"],
)

_default_detail_fields = (
    keypair_resource_policy_fields["name"],
    keypair_resource_policy_fields["created_at"],
    keypair_resource_policy_fields["total_resource_slots"],
    keypair_resource_policy_fields["max_concurrent_sessions"],
    keypair_resource_policy_fields["max_vfolder_count"],
    keypair_resource_policy_fields["max_vfolder_size"],
    keypair_resource_policy_fields["idle_timeout"],
    keypair_resource_policy_fields["max_containers_per_session"],
    keypair_resource_policy_fields["allowed_vfolder_hosts"],
)


class KeypairResourcePolicy(BaseFunction):
    """
    Provides interactions with keypair resource policy.
    """

    def __init__(self, access_key: str):
        self.access_key = access_key

    @api_function
    @classmethod
    async def create(
        cls,
        name: str,
        default_for_unspecified: int,
        total_resource_slots: str,
        max_session_lifetime: int,
        max_concurrent_sessions: int,
        max_containers_per_session: int,
        max_vfolder_count: int,
        max_vfolder_size: int,
        idle_timeout: int,
        allowed_vfolder_hosts: Optional[str] = None,
        fields: Iterable[FieldSpec | str] = None,
    ) -> dict:
        """
        Creates a new keypair resource policy with the given options.
        You need an admin privilege for this operation.
        """
        q = (
            "mutation($name: String!, $input: CreateKeyPairResourcePolicyInput!) {"
            + "  create_keypair_resource_policy(name: $name, props: $input) {"
            "    ok msg resource_policy { $fields }"
            "  }"
            "}"
        )
        resolved_fields = resolve_fields(
            fields, keypair_resource_policy_fields, (keypair_resource_policy_fields["name"],)
        )
        q = q.replace("$fields", " ".join(resolved_fields))
        variables = {
            "name": name,
            "input": {
                "default_for_unspecified": default_for_unspecified,
                "total_resource_slots": total_resource_slots,
                "max_session_lifetime": max_session_lifetime,
                "max_concurrent_sessions": max_concurrent_sessions,
                "max_containers_per_session": max_containers_per_session,
                "max_vfolder_count": max_vfolder_count,
                "max_vfolder_size": max_vfolder_size,
                "idle_timeout": idle_timeout,
                "allowed_vfolder_hosts": allowed_vfolder_hosts,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["create_keypair_resource_policy"]

    @api_function
    @classmethod
    async def update(
        cls,
        name: str,
        default_for_unspecified: int,
        max_session_lifetime: int,
        max_concurrent_sessions: int,
        max_containers_per_session: int,
        max_vfolder_count: int,
        max_vfolder_size: int,
        idle_timeout: int,
        total_resource_slots: Optional[str] = None,
        allowed_vfolder_hosts: Optional[str] = None,
    ) -> dict:
        """
        Updates an existing keypair resource policy with the given options.
        You need an admin privilege for this operation.
        """
        q = (
            "mutation($name: String!, $input: ModifyKeyPairResourcePolicyInput!) {"
            + "  modify_keypair_resource_policy(name: $name, props: $input) {"
            "    ok msg"
            "  }"
            "}"
        )
        variables = {
            "name": name,
            "input": {
                "default_for_unspecified": default_for_unspecified,
                "total_resource_slots": total_resource_slots,
                "max_session_lifetime": max_session_lifetime,
                "max_concurrent_sessions": max_concurrent_sessions,
                "max_containers_per_session": max_containers_per_session,
                "max_vfolder_count": max_vfolder_count,
                "max_vfolder_size": max_vfolder_size,
                "idle_timeout": idle_timeout,
                "allowed_vfolder_hosts": allowed_vfolder_hosts,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["modify_keypair_resource_policy"]

    @api_function
    @classmethod
    async def delete(cls, name: str) -> dict:
        """
        Deletes an existing keypair resource policy with given name.
        You need an admin privilege for this operation.
        """
        q = (
            "mutation($name: String!) {" + "  delete_keypair_resource_policy(name: $name) {"
            "    ok msg"
            "  }"
            "}"
        )
        variables = {
            "name": name,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["delete_keypair_resource_policy"]

    @api_function
    @classmethod
    async def list(
        cls,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Lists the keypair resource policies.
        You need an admin privilege for this operation.
        """
        q = "query {" "  keypair_resource_policies {" "    $fields" "  }" "}"
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(q)
        return data["keypair_resource_policies"]

    @api_function
    async def info(
        self,
        name: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Returns the resource policy's information.

        :param fields: Additional per-agent query fields to fetch.

        .. versionadded:: 19.03
        """
        q = (
            "query($name: String) {"
            "  keypair_resource_policy(name: $name) {"
            "    $fields"
            "  }"
            "}"
        )
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {
            "name": name,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["keypair_resource_policy"]



import json
import re
from decimal import Decimal
from typing import Any, Mapping, Optional, Union

import click


class ByteSizeParamType(click.ParamType):
    name = "byte"

    _rx_digits = re.compile(r"^(\d+(?:\.\d*)?)([kmgtpe]?)$", re.I)
    _scales = {
        "k": 2**10,
        "m": 2**20,
        "g": 2**30,
        "t": 2**40,
        "p": 2**50,
        "e": 2**60,
    }

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value
        if not isinstance(value, str):
            self.fail(
                f"expected string, got {value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        m = self._rx_digits.search(value)
        if m is None:
            self.fail(f"{value!r} is not a valid byte-size expression", param, ctx)
        size = float(m.group(1))
        unit = m.group(2).lower()
        return int(size * self._scales.get(unit, 1))


class ByteSizeParamCheckType(ByteSizeParamType):
    name = "byte-check"

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value
        if not isinstance(value, str):
            self.fail(
                f"expected string, got {value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        m = self._rx_digits.search(value)
        if m is None:
            self.fail(f"{value!r} is not a valid byte-size expression", param, ctx)
        return value


class CommaSeparatedKVListParamType(click.ParamType):
    name = "comma-seperated-KVList-check"

    def convert(self, value: Union[str, Mapping[str, str]], param, ctx) -> Mapping[str, str]:
        if isinstance(value, dict):
            return value
        if not isinstance(value, str):
            self.fail(
                f"expected string, got {value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        override_map = {}
        for assignment in value.split(","):
            try:
                k, _, v = assignment.partition("=")
                if k == "" or v == "":
                    raise ValueError(f"key or value is empty. key = {k}, value = {v}")
            except ValueError:
                self.fail(
                    f"{value!r} is not a valid mapping expression",
                    param,
                    ctx,
                )
            else:
                override_map[k] = v
        return override_map


class JSONParamType(click.ParamType):
    """
    A JSON string parameter type.
    The default value must be given as a valid JSON-parsable string,
    not the Python objects.
    """

    name = "json-string"

    def __init__(self) -> None:
        super().__init__()
        self._parsed = False

    def convert(
        self,
        value: Optional[str],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Any:
        if self._parsed:
            # Click invokes this method TWICE
            # for a default value given as string.
            return value
        self._parsed = True
        if value is None:
            return None
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            self.fail(f"cannot parse {value!r} as JSON", param, ctx)
        return value


def drange(start: Decimal, stop: Decimal, num: int):
    """
    A simplified version of numpy.linspace with default options
    """
    delta = stop - start
    step = delta / (num - 1)
    yield from (start + step * Decimal(tick) for tick in range(0, num))


class RangeExprOptionType(click.ParamType):
    """
    Accepts a range expression which generates a range of values for a variable.

    Linear space range: "linspace:1,2,10" (start, stop, num) as in numpy.linspace
    Pythonic range: "range:1,10,2" (start, stop[, step]) as in Python's range
    Case range: "case:a,b,c" (comma-separated strings)
    """

    _rx_range_key = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*$")
    name = "Range Expression"

    def convert(self, arg, param, ctx):
        key, value = arg.split("=", maxsplit=1)
        assert self._rx_range_key.match(key), "The key must be a valid slug string."
        try:
            if value.startswith("case:"):
                return key, value[5:].split(",")
            elif value.startswith("linspace:"):
                start, stop, num = value[9:].split(",")
                return key, tuple(drange(Decimal(start), Decimal(stop), int(num)))
            elif value.startswith("range:"):
                range_args = map(int, value[6:].split(","))
                return key, tuple(range(*range_args))
            else:
                self.fail("Unrecognized range expression type", param, ctx)
        except ValueError as e:
            self.fail(str(e), param, ctx)


class CommaSeparatedListType(click.ParamType):

    name = "List Expression"

    def convert(self, arg, param, ctx):
        try:
            if isinstance(arg, int):
                return arg
            elif isinstance(arg, str):
                return arg.split(",")
        except ValueError as e:
            self.fail(repr(e), param, ctx)



import json
import sys
from datetime import datetime
from pathlib import Path

import click
import humanize
from tabulate import tabulate
from tqdm import tqdm

from ai.backend.cli.interaction import ask_yn
from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode
from ai.backend.client.config import DEFAULT_CHUNK_SIZE, APIConfig
from ai.backend.client.session import Session

from ..compat import asyncio_run
from ..session import AsyncSession
from .params import ByteSizeParamCheckType, ByteSizeParamType, CommaSeparatedKVListParamType
from .pretty import print_done, print_error, print_fail, print_info, print_wait, print_warn


@main.group()
def vfolder():
    """Set of vfolder operations"""


@vfolder.command()
def list_hosts():
    """List the hosts of virtual folders that is accessible to the current user."""
    with Session() as session:
        try:
            resp = session.VFolder.list_hosts()
            print("Default vfolder host: {}".format(resp["default"]))
            print("Usable hosts: {}".format(", ".join(resp["allowed"])))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
def list_allowed_types():
    """List allowed vfolder types."""
    with Session() as session:
        try:
            resp = session.VFolder.list_allowed_types()
            print(resp)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("host", type=str, default=None)
@click.option(
    "-g",
    "--group",
    metavar="GROUP",
    type=str,
    default=None,
    help="Group ID or NAME. Specify this option if you want to create a group folder.",
)
@click.option(
    "--unmanaged",
    "host_path",
    type=bool,
    is_flag=True,
    help="Treats HOST as a mount point of unmanaged virtual folder. "
    "This option can only be used by Admin or Superadmin.",
)
@click.option(
    "-m",
    "--usage-mode",
    metavar="USAGE_MODE",
    type=str,
    default="general",
    help='Purpose of the folder. Normal folders are usually set to "general". '
    'Available options: "general", "data" (provides data to users), '
    'and "model" (provides pre-trained models).',
)
@click.option(
    "-p",
    "--permission",
    metavar="PERMISSION",
    type=str,
    default="rw",
    help="Folder's innate permission. "
    'Group folders can be shared as read-only by setting this option to "ro".'
    "Invited folders override this setting by its own invitation permission.",
)
@click.option(
    "-q",
    "--quota",
    metavar="QUOTA",
    type=ByteSizeParamCheckType(),
    default="0",
    help="Quota of the virtual folder. "
    "(Use 'm' for megabytes, 'g' for gigabytes, and etc.) "
    "Default is maximum amount possible.",
)
@click.option(
    "--cloneable",
    "--allow-clone",
    type=bool,
    is_flag=True,
    help="Allows the virtual folder to be cloned by users.",
)
def create(name, host, group, host_path, usage_mode, permission, quota, cloneable):
    """Create a new virtual folder.

    \b
    NAME: Name of a virtual folder.
    HOST: Name of a virtual folder host in which the virtual folder will be created.
    """
    with Session() as session:
        try:
            if host_path:
                result = session.VFolder.create(
                    name=name,
                    unmanaged_path=host,
                    group=group,
                    usage_mode=usage_mode,
                    permission=permission,
                    quota=quota,
                    cloneable=cloneable,
                )
            else:
                result = session.VFolder.create(
                    name=name,
                    host=host,
                    group=group,
                    usage_mode=usage_mode,
                    permission=permission,
                    quota=quota,
                    cloneable=cloneable,
                )
            print('Virtual folder "{0}" is created.'.format(result["name"]))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
def delete(name):
    """Delete the given virtual folder. This operation is irreversible!

    NAME: Name of a virtual folder.
    """
    with Session() as session:
        try:
            session.VFolder(name).delete()
            print_done("Deleted.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("old_name", type=str)
@click.argument("new_name", type=str)
def rename(old_name, new_name):
    """Rename the given virtual folder. This operation is irreversible!
    You cannot change the vfolders that are shared by other users,
    and the new name must be unique among all your accessible vfolders
    including the shared ones.

    OLD_NAME: The current name of a virtual folder.
    NEW_NAME: The new name of a virtual folder.
    """
    with Session() as session:
        try:
            session.VFolder(old_name).rename(new_name)
            print_done("Renamed.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
def info(name):
    """Show the information of the given virtual folder.

    NAME: Name of a virtual folder.
    """
    with Session() as session:
        try:
            result = session.VFolder(name).info()
            print('Virtual folder "{0}" (ID: {1})'.format(result["name"], result["id"]))
            print("- Owner:", result["is_owner"])
            print("- Permission:", result["permission"])
            print("- Status: {0}".format(result["status"]))
            print("- Number of files: {0}".format(result["numFiles"]))
            print("- Ownership Type: {0}".format(result["type"]))
            print("- Permission:", result["permission"])
            print("- Usage Mode: {0}".format(result.get("usage_mode", "")))
            print("- Group ID: {0}".format(result["group"]))
            print("- User ID: {0}".format(result["user"]))
            print("- Clone Allowed: {0}".format(result["cloneable"]))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command(context_settings={"show_default": True})  # bug: pallets/click#1565 (fixed in 8.0)
@click.argument("name", type=str)
@click.argument("filenames", type=Path, nargs=-1)
@click.option(
    "-b",
    "--base-dir",
    type=Path,
    default=None,
    help="Set the parent directory from where the file is uploaded.  "
    "[default: current working directry]",
)
@click.option(
    "--chunk-size",
    type=ByteSizeParamType(),
    default=humanize.naturalsize(DEFAULT_CHUNK_SIZE, binary=True, gnu=True),
    help='Transfer the file with the given chunk size with binary suffixes (e.g., "16m"). '
    "Set this between 8 to 64 megabytes for high-speed disks (e.g., SSD RAID) "
    "and networks (e.g., 40 GbE) for the maximum throughput.",
)
@click.option(
    "--override-storage-proxy",
    type=CommaSeparatedKVListParamType(),
    default=None,
    help="Overrides storage proxy address. "
    'The value must shape like "X1=Y1,X2=Y2...". '
    "Each Yn address must at least include the IP address "
    "or the hostname and may include the protocol part and the port number to replace.",
)
def upload(name, filenames, base_dir, chunk_size, override_storage_proxy):
    """
    TUS Upload a file to the virtual folder from the current working directory.
    The files with the same names will be overwirtten.

    \b
    NAME: Name of a virtual folder.
    FILENAMES: Paths of the files to be uploaded.
    """
    with Session() as session:
        try:
            session.VFolder(name).upload(
                filenames,
                basedir=base_dir,
                chunk_size=chunk_size,
                show_progress=True,
                address_map=override_storage_proxy
                or APIConfig.DEFAULTS["storage_proxy_address_map"],
            )
            print_done("Done.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command(context_settings={"show_default": True})  # bug: pallets/click#1565 (fixed in 8.0)
@click.argument("name", type=str)
@click.argument("filenames", type=Path, nargs=-1)
@click.option(
    "-b",
    "--base-dir",
    type=Path,
    default=None,
    help="Set the parent directory from where the file is uploaded.  "
    "[default: current working directry]",
)
@click.option(
    "--chunk-size",
    type=ByteSizeParamType(),
    default=humanize.naturalsize(DEFAULT_CHUNK_SIZE, binary=True, gnu=True),
    help='Transfer the file with the given chunk size with binary suffixes (e.g., "16m"). '
    "Set this between 8 to 64 megabytes for high-speed disks (e.g., SSD RAID) "
    "and networks (e.g., 40 GbE) for the maximum throughput.",
)
@click.option(
    "--override-storage-proxy",
    type=CommaSeparatedKVListParamType(),
    default=None,
    help="Overrides storage proxy address. "
    'The value must shape like "X1=Y1,X2=Y2...". '
    "Each Yn address must at least include the IP address "
    "or the hostname and may include the protocol part and the port number to replace.",
)
@click.option(
    "--max-retries",
    type=int,
    default=20,
    help="Maximum retry attempt when any failure occurs.",
)
def download(name, filenames, base_dir, chunk_size, override_storage_proxy, max_retries):
    """
    Download a file from the virtual folder to the current working directory.
    The files with the same names will be overwirtten.

    \b
    NAME: Name of a virtual folder.
    FILENAMES: Paths of the files to be downloaded inside a vfolder.
    """
    with Session() as session:
        try:
            session.VFolder(name).download(
                filenames,
                basedir=base_dir,
                chunk_size=chunk_size,
                show_progress=True,
                address_map=override_storage_proxy
                or APIConfig.DEFAULTS["storage_proxy_address_map"],
                max_retries=max_retries,
            )
            print_done("Done.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("filename", type=Path)
def request_download(name, filename):
    """
    Request JWT-formated download token for later use.

    \b
    NAME: Name of a virtual folder.
    FILENAME: Path of the file to be downloaded.
    """
    with Session() as session:
        try:
            response = json.loads(session.VFolder(name).request_download(filename))
            print_done(f'Download token: {response["token"]}')
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("filenames", nargs=-1)
def cp(filenames):
    """An scp-like shortcut for download/upload commands.

    FILENAMES: Paths of the files to operate on. The last one is the target while all
               others are the sources.  Either source paths or the target path should
               be prefixed with "<vfolder-name>:" like when using the Linux scp
               command to indicate if it is a remote path.
    """
    raise NotImplementedError


@vfolder.command()
@click.argument("name", type=str)
@click.argument("path", type=str)
@click.option(
    "-p",
    "--parents",
    default=False,
    is_flag=True,
    help="Make missing parents of this path as needed",
)
@click.option(
    "-e", "--exist-ok", default=False, is_flag=True, help="Skip an error caused by file not found"
)
def mkdir(name, path, parents, exist_ok):
    """Create an empty directory in the virtual folder.

    \b
    NAME: Name of a virtual folder.
    PATH: The name or path of directory. Parent directories are created automatically
          if they do not exist.
    """
    with Session() as session:
        try:
            session.VFolder(name).mkdir(path, parents=parents, exist_ok=exist_ok)
            print_done("Done.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("target_path", type=str)
@click.argument("new_name", type=str)
def rename_file(name, target_path, new_name):
    """
    Rename a file or a directory in a virtual folder.

    \b
    NAME: Name of a virtual folder.
    TARGET_PATH: The target path inside a virtual folder (file or directory).
    NEW_NAME: New name of the target (should not contain slash).
    """
    with Session() as session:
        try:
            session.VFolder(name).rename_file(target_path, new_name)
            print_done("Renamed.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("src", type=str)
@click.argument("dst", type=str)
def mv(name, src, dst):
    """
    Move a file or a directory within a virtual folder.
    If the destination is a file and already exists, it will be overwritten.
    If the destination is a directory, the source file or directory
    is moved inside it.

    \b
    NAME: Name of a virtual folder.
    SRC: The relative path of the source file or directory inside a virtual folder
    DST: The relative path of the destination file or directory inside a virtual folder.
    """
    with Session() as session:
        try:
            session.VFolder(name).move_file(src, dst)
            print_done("Moved.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command(aliases=["delete-file"])
@click.argument("name", type=str)
@click.argument("filenames", nargs=-1)
@click.option("-r", "--recursive", is_flag=True, help="Enable recursive deletion of directories.")
def rm(name, filenames, recursive):
    """
    Delete files in a virtual folder.
    If one of the given paths is a directory and the recursive option is enabled,
    all its content and the directory itself are recursively deleted.

    This operation is irreversible!

    \b
    NAME: Name of a virtual folder.
    FILENAMES: Paths of the files to delete.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info("Cancelled")
                sys.exit(ExitCode.FAILURE)
            session.VFolder(name).delete_files(filenames, recursive=recursive)
            print_done("Done.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("path", metavar="PATH", nargs=1, default=".")
def ls(name, path):
    """
    List files in a path of a virtual folder.

    \b
    NAME: Name of a virtual folder.
    PATH: Path inside vfolder.
    """
    with Session() as session:
        try:
            print_wait('Retrieving list of files in "{}"...'.format(path))
            result = session.VFolder(name).list_files(path)
            if "error_msg" in result and result["error_msg"]:
                print_fail(result["error_msg"])
                return
            files = json.loads(result["files"])
            table = []
            headers = ["file name", "size", "modified", "mode"]
            for file in files:
                mdt = datetime.fromtimestamp(file["mtime"])
                mtime = mdt.strftime("%b %d %Y %H:%M:%S")
                row = [file["filename"], file["size"], mtime, file["mode"]]
                table.append(row)
            print_done("Retrived.")
            print(tabulate(table, headers=headers))
        except Exception as e:
            print_error(e)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("emails", type=str, nargs=-1, required=True)
@click.option(
    "-p",
    "--perm",
    metavar="PERMISSION",
    type=str,
    default="rw",
    help='Permission to give. "ro" (read-only) / "rw" (read-write) / "wd" (write-delete).',
)
def invite(name, emails, perm):
    """Invite other users to access a user-type virtual folder.

    \b
    NAME: Name of a virtual folder.
    EMAILS: Emails to invite.
    """
    with Session() as session:
        try:
            assert perm in ["rw", "ro", "wd"], "Invalid permission: {}".format(perm)
            result = session.VFolder(name).invite(perm, emails)
            invited_ids = result.get("invited_ids", [])
            if invited_ids:
                print("Invitation sent to:")
                for invitee in invited_ids:
                    print("\t- " + invitee)
            else:
                print("No users found. Invitation was not sent.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
def invitations():
    """List and manage received invitations."""
    with Session() as session:
        try:
            result = session.VFolder.invitations()
            invitations = result.get("invitations", [])
            if len(invitations) < 1:
                print("No invitations.")
                return
            print("List of invitations (inviter, vfolder id, permission):")
            for cnt, inv in enumerate(invitations):
                if inv["perm"] == "rw":
                    perm = "read-write"
                elif inv["perm"] == "ro":
                    perm = "read-only"
                else:
                    perm = inv["perm"]
                print("[{}] {}, {}, {}".format(cnt + 1, inv["inviter"], inv["vfolder_id"], perm))

            selection = input("Choose invitation number to manage: ")
            if selection.isdigit():
                selection = int(selection) - 1
            else:
                return
            if 0 <= selection < len(invitations):
                while True:
                    action = input("Choose action. (a)ccept, (r)eject, (c)ancel: ")
                    if action.lower() == "a":
                        session.VFolder.accept_invitation(invitations[selection]["id"])
                        msg = "You can now access vfolder {} ({})".format(
                            invitations[selection]["vfolder_name"],
                            invitations[selection]["id"],
                        )
                        print(msg)
                        break
                    elif action.lower() == "r":
                        session.VFolder.delete_invitation(invitations[selection]["id"])
                        msg = "vfolder invitation rejected: {} ({})".format(
                            invitations[selection]["vfolder_name"],
                            invitations[selection]["id"],
                        )
                        print(msg)
                        break
                    elif action.lower() == "c":
                        break
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("emails", type=str, nargs=-1, required=True)
@click.option(
    "-p",
    "--perm",
    metavar="PERMISSION",
    type=str,
    default="rw",
    help='Permission to give. "ro" (read-only) / "rw" (read-write) / "wd" (write-delete).',
)
def share(name, emails, perm):
    """Share a group folder to users with overriding permission.

    \b
    NAME: Name of a (group-type) virtual folder.
    EMAILS: Emails to share.
    """
    with Session() as session:
        try:
            assert perm in ["rw", "ro", "wd"], "Invalid permission: {}".format(perm)
            result = session.VFolder(name).share(perm, emails)
            shared_emails = result.get("shared_emails", [])
            if shared_emails:
                print("Shared with {} permission to:".format(perm))
                for _email in shared_emails:
                    print("\t- " + _email)
            else:
                print("No users found. Folder is not shared.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("emails", type=str, nargs=-1, required=True)
def unshare(name, emails):
    """Unshare a group folder from users.

    \b
    NAME: Name of a (group-type) virtual folder.
    EMAILS: Emails to share.
    """
    with Session() as session:
        try:
            result = session.VFolder(name).unshare(emails)
            unshared_emails = result.get("unshared_emails", [])
            if unshared_emails:
                print("Unshared from:")
                for _email in unshared_emails:
                    print("\t- " + _email)
            else:
                print("No users found. Folder is not unshared.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.option(
    "-s",
    "--shared-user-uuid",
    metavar="SHARED_USER_UUID",
    type=str,
    default=None,
    help="The ID of the person who wants to leave (the person who shared the vfolder).",
)
def leave(name, shared_user_uuid):
    """Leave the shared virutal folder.

    NAME: Name of a virtual folder
    """
    with Session() as session:
        try:
            vfolder_info = session.VFolder(name).info()
            if vfolder_info["type"] == "group":
                print("You cannot leave a group virtual folder.")
                return
            if vfolder_info["is_owner"]:
                print("You cannot leave a virtual folder you own. Consider using delete instead.")
                return
            session.VFolder(name).leave(shared_user_uuid)
            print('Left the shared virtual folder "{}".'.format(name))

        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@vfolder.command()
@click.argument("name", type=str)
@click.argument("target_name", type=str)
@click.argument("target_host", type=str)
@click.option(
    "-m",
    "--usage-mode",
    metavar="USAGE_MODE",
    type=str,
    default="general",
    help="Purpose of the cloned virtual folder. " "Default value is 'general'.",
)
@click.option(
    "-p",
    "--permission",
    metavar="PERMISSION",
    type=str,
    default="rw",
    help="Cloned virtual folder's permission. " "Default value is 'rw'.",
)
def clone(name, target_name, target_host, usage_mode, permission):
    """Clone a virtual folder.

    \b
    NAME: Name of the virtual folder to clone from.
    TARGET_NAME: Name of the virtual folder to clone to.
    TARGET_HOST: Name of a virtual folder host to which the virtual folder will be cloned.
    """
    with Session() as session:
        try:
            vfolder_info = session.VFolder(name).info()
            if not vfolder_info["cloneable"]:
                print(
                    "Clone is not allowed for this virtual folder. "
                    "Please update the 'cloneable' option."
                )
                return
            result = session.VFolder(name).clone(
                target_name,
                target_host=target_host,
                usage_mode=usage_mode,
                permission=permission,
            )
            bgtask_id = result.get("bgtask_id")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)

    async def clone_vfolder_tracker(bgtask_id):
        print_wait(
            "Cloning the vfolder... "
            "(This may take a while depending on its size and number of files!)",
        )
        async with AsyncSession() as session:
            try:
                bgtask = session.BackgroundTask(bgtask_id)
                completion_msg_func = lambda: print_done("Cloning the vfolder is complete.")
                async with bgtask.listen_events() as response:
                    # TODO: get the unit of progress from response
                    with tqdm(unit="bytes", disable=True) as pbar:
                        async for ev in response:
                            data = json.loads(ev.data)
                            if ev.event == "bgtask_updated":
                                pbar.total = data["total_progress"]
                                pbar.write(data["message"])
                                pbar.update(data["current_progress"] - pbar.n)
                            elif ev.event == "bgtask_failed":
                                error_msg = data["message"]
                                completion_msg_func = lambda: print_fail(
                                    f"Error during the operation: {error_msg}",
                                )
                            elif ev.event == "bgtask_cancelled":
                                completion_msg_func = lambda: print_warn(
                                    "The operation has been cancelled in the middle. "
                                    "(This may be due to server shutdown.)",
                                )
            finally:
                completion_msg_func()

    if bgtask_id is None:
        print_done("Cloning the vfolder is complete.")
    else:
        asyncio_run(clone_vfolder_tracker(bgtask_id))


@vfolder.command()
@click.argument("name", type=str)
@click.option(
    "-p", "--permission", type=str, metavar="PERMISSION", help="Folder's innate permission."
)
@click.option(
    "--set-cloneable",
    type=bool,
    metavar="BOOLEXPR",
    help="A boolean-interpretable string whether a virtual folder can be cloned. "
    "If not set, the cloneable property is not changed.",
)
def update_options(name, permission, set_cloneable):
    """Update an existing virtual folder.

    \b
    NAME: Name of the virtual folder to update.
    """
    with Session() as session:
        try:
            vfolder_info = session.VFolder(name).info()
            if not vfolder_info["is_owner"]:
                print("You cannot update virtual folder that you do not own.")
                return
            session.VFolder(name).update_options(
                name,
                permission=permission,
                cloneable=set_cloneable,
            )
            print_done("Updated.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)



import sys

import click

from ai.backend.cli.types import ExitCode
from ai.backend.client.output.fields import keypair_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..types import CLIContext
from . import admin


@admin.group()
def keypair() -> None:
    """
    KeyPair administration commands.
    """


@keypair.command()
@pass_ctx_obj
def info(ctx: CLIContext) -> None:
    """
    Show the server-side information of the currently configured access key.
    """
    fields = [
        keypair_fields["user_id"],
        keypair_fields["full_name"],
        keypair_fields["access_key"],
        keypair_fields["secret_key"],
        keypair_fields["is_active"],
        keypair_fields["is_admin"],
        keypair_fields["created_at"],
        keypair_fields["last_used"],
        keypair_fields["resource_policy"],
        keypair_fields["rate_limit"],
        keypair_fields["concurrency_used"],
    ]
    with Session() as session:
        try:
            kp = session.KeyPair(session.config.access_key)
            item = kp.info(fields=fields)
            ctx.output.print_item(item, fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@keypair.command()
@pass_ctx_obj
@click.option(
    "-u",
    "--user-id",
    type=str,
    default=None,
    help="Show keypairs of this given user. [default: show all]",
)
@click.option("--is-active", type=bool, default=None, help="Filter keypairs by activation.")
@click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
@click.option("--order", default=None, help="Set the query ordering expression.")
@click.option("--offset", default=0, help="The index of the current page start for pagination.")
@click.option("--limit", type=int, default=None, help="The page size for pagination.")
def list(ctx: CLIContext, user_id, is_active, filter_, order, offset, limit) -> None:
    """
    List keypairs.
    To show all keypairs or other user's, your access key must have the admin
    privilege.
    (admin privilege required)
    """
    fields = [
        keypair_fields["user_id"],
        keypair_fields["full_name"],
        keypair_fields["access_key"],
        keypair_fields["secret_key"],
        keypair_fields["is_active"],
        keypair_fields["is_admin"],
        keypair_fields["created_at"],
        keypair_fields["last_used"],
        keypair_fields["resource_policy"],
        keypair_fields["rate_limit"],
        keypair_fields["concurrency_used"],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.KeyPair.paginated_list(
                is_active,
                user_id=user_id,
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(ExitCode.FAILURE)


@keypair.command()
@pass_ctx_obj
@click.argument("user-id", type=str, default=None, metavar="USERID")
@click.argument("resource-policy", type=str, default=None, metavar="RESOURCE_POLICY")
@click.option("-a", "--admin", is_flag=True, help="Give the admin privilege to the new keypair.")
@click.option("-i", "--inactive", is_flag=True, help="Create the new keypair in inactive state.")
@click.option("-r", "--rate-limit", type=int, default=5000, help="Set the API query rate limit.")
def add(ctx: CLIContext, user_id, resource_policy, admin, inactive, rate_limit):
    """
    Add a new keypair.

    USER_ID: User ID of a new key pair.
    RESOURCE_POLICY: resource policy for new key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.create(
                user_id,
                is_active=not inactive,
                is_admin=admin,
                resource_policy=resource_policy,
                rate_limit=rate_limit,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="keypair",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="keypair",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            item_name="keypair",
            extra_info={
                "access_key": data["keypair"]["access_key"],
                "secret_key": data["keypair"]["secret_key"],
            },
        )


@keypair.command()
@pass_ctx_obj
@click.argument("access_key", type=str, default=None, metavar="ACCESSKEY")
@click.option("--resource-policy", type=str, help="Resource policy for the keypair.")
@click.option("--is-admin", type=bool, help="Set admin privilege.")
@click.option("--is-active", type=bool, help="Set key pair active or not.")
@click.option("-r", "--rate-limit", type=int, help="Set the API query rate limit.")
def update(ctx: CLIContext, access_key, resource_policy, is_admin, is_active, rate_limit):
    """
    Update an existing keypair.

    ACCESS_KEY: Access key of an existing key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.update(
                access_key,
                is_active=is_active,
                is_admin=is_admin,
                resource_policy=resource_policy,
                rate_limit=rate_limit,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="keypair",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="keypair",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "access_key": access_key,
            },
        )


@keypair.command()
@pass_ctx_obj
@click.argument("access-key", type=str, metavar="ACCESSKEY")
def delete(ctx: CLIContext, access_key):
    """
    Delete an existing keypair.

    ACCESSKEY: ACCESSKEY for a keypair to delete.
    """
    with Session() as session:
        try:
            data = session.KeyPair.delete(access_key)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="keypair",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="keypair",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "access_key": access_key,
            },
        )


@keypair.command()
@pass_ctx_obj
@click.argument("access-key", type=str, metavar="ACCESSKEY")
def activate(ctx: CLIContext, access_key):
    """
    Activate an inactivated keypair.

    ACCESS_KEY: Access key of an existing key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.activate(access_key)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="keypair",
                action_name="activation",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="keypair",
                action_name="activation",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "access_key": access_key,
            },
        )


@keypair.command()
@pass_ctx_obj
@click.argument("access-key", type=str, metavar="ACCESSKEY")
def deactivate(ctx: CLIContext, access_key):
    """
    Deactivate an active keypair.

    ACCESS_KEY: Access key of an existing key pair.
    """
    with Session() as session:
        try:
            data = session.KeyPair.deactivate(access_key)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="keypair",
                action_name="deactivation",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="keypair",
                action_name="deactivation",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "access_key": access_key,
            },
        )



import sys

import click
from tabulate import tabulate

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode

from ..session import Session
from .pretty import print_error, print_info, print_warn


@main.group(aliases=["sesstpl"])
def session_template():
    """Set of session template operations"""


@session_template.command()
@click.option(
    "-f",
    "--file",
    "template_path",
    help="Path to task template file. "
    "If not specified, client will try to read config from STDIN. ",
)
@click.option(
    "-d",
    "--domain",
    metavar="DOMAIN_NAME",
    default=None,
    help="Domain name where the session will be spawned. "
    "If not specified, config's domain name will be used.",
)
@click.option(
    "-g",
    "--group",
    metavar="GROUP_NAME",
    default=None,
    help="Group name where the session is spawned. "
    "User should be a member of the group to execute the code.",
)
@click.option(
    "-o",
    "--owner",
    "--owner-access-key",
    "owner_access_key",
    metavar="ACCESS_KEY",
    help="Set the owner of the target session explicitly.",
)
def create(template_path, domain, group, owner_access_key):
    """
    Store task template to Backend.AI Manager and return template ID.
    Template can be used when creating new session.
    """

    if template_path:
        with open(template_path, "r") as fr:
            body = fr.read()
    else:
        body = ""
        for line in sys.stdin:
            body += line + "\n"
    with Session() as session:
        try:
            # TODO: Make user select template type when cluster template is implemented
            template = session.SessionTemplate.create(
                body, domain_name=domain, group_name=group, owner_access_key=owner_access_key
            )
            print_info(f"Task template {template.template_id} created and ready")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session_template.command()
@click.argument("template_id", metavar="TEMPLATEID")
@click.option(
    "-f",
    "--format",
    "template_format",
    default="yaml",
    help='Output format for task template. "yaml" and "json" allowed.',
)
@click.option(
    "-o",
    "--owner",
    "--owner-access-key",
    "owner_access_key",
    metavar="ACCESS_KEY",
    help="Specify the owner of the target session explicitly.",
)
def get(template_id, template_format, owner_access_key):
    """
    Print task template associated with given template ID
    """
    with Session() as session:
        try:
            template = session.SessionTemplate(template_id, owner_access_key=owner_access_key)
            body = template.get(body_format=template_format)
            print(body)
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session_template.command()
@click.option(
    "-a",
    "--list-all",
    is_flag=True,
    help="List all virtual folders (superadmin privilege is required).",
)
def list(list_all):
    """
    List all availabe task templates by user.
    """
    fields = [
        ("Name", "name"),
        ("ID", "id"),
        ("Created At", "created_at"),
        ("Owner", "is_owner"),
        ("Type", "type"),
        ("User", "user"),
        ("Group", "group"),
    ]
    with Session() as session:
        try:
            resp = session.SessionTemplate.list_templates(list_all)
            if not resp:
                print("There is no task templates created yet.")
                return
            rows = (tuple(vf[key] for _, key in fields) for vf in resp)
            hdrs = (display_name for display_name, _ in fields)
            print(tabulate(rows, hdrs))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session_template.command()
@click.argument("template_id", metavar="TEMPLATEID")
@click.option(
    "-f",
    "--file",
    "template_path",
    help="Path to task template file. "
    "If not specified, client will try to read config from STDIN. ",
)
@click.option(
    "-o",
    "--owner",
    "--owner-access-key",
    "owner_access_key",
    metavar="ACCESS_KEY",
    help="Specify the owner of the target session explicitly.",
)
def update(template_id, template_path, owner_access_key):
    """
    Update task template stored in Backend.AI Manager.
    """

    if template_path:
        with open(template_path, "r") as fr:
            body = fr.read()
    else:
        body = ""
        for line in sys.stdin:
            body += line + "\n"
    with Session() as session:
        try:
            template = session.SessionTemplate(template_id, owner_access_key=owner_access_key)
            template.put(body)
            print_info(f"Task template {template.template_id} updated")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session_template.command()
@click.argument("template_id", metavar="TEMPLATEID")
@click.option(
    "-f",
    "--force",
    type=bool,
    is_flag=True,
    help="If specified, delete task template without asking.",
)
@click.option(
    "-o",
    "--owner",
    "--owner-access-key",
    "owner_access_key",
    metavar="ACCESS_KEY",
    help="Specify the owner of the target session explicitly.",
)
def delete(template_id, force, owner_access_key):
    """
    Delete task template from Backend.AI Manager.
    """
    with Session() as session:
        template = session.SessionTemplate(template_id, owner_access_key=owner_access_key)
        if not force:
            print_warn("Are you sure? (y/[n])")
            result = input()
            if result.strip() != "y":
                print_info("Aborting.")
                exit()
        try:
            template.delete()
            print_info(f"Task template {template.template_id} deleted")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)



from __future__ import annotations

import asyncio
import json
import secrets
import subprocess
import sys
import uuid
from datetime import datetime, timedelta
from pathlib import Path
from typing import IO, List, Literal, Optional, Sequence

import click
import inquirer
from async_timeout import timeout
from dateutil.parser import isoparse
from dateutil.tz import tzutc
from humanize import naturalsize
from tabulate import tabulate

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode

from ..compat import asyncio_run
from ..exceptions import BackendAPIError
from ..func.session import ComputeSession
from ..output.fields import session_fields
from ..output.types import FieldSpec
from ..session import AsyncSession, Session
from ..types import Undefined, undefined
from . import events
from .params import CommaSeparatedListType
from .pretty import print_done, print_error, print_fail, print_info, print_wait, print_warn
from .run import format_stats, prepare_env_arg, prepare_mount_arg, prepare_resource_arg
from .ssh import container_ssh_ctx

list_expr = CommaSeparatedListType()


@main.group()
def session():
    """Set of compute session operations"""


def _create_cmd(docs: str = None):
    @click.argument("image")
    @click.option(
        "-t",
        "--name",
        "--client-token",
        metavar="NAME",
        help="Specify a human-readable session name. " "If not set, a random hex string is used.",
    )
    @click.option(
        "-o",
        "--owner",
        "--owner-access-key",
        metavar="ACCESS_KEY",
        help="Set the owner of the target session explicitly.",
    )
    # job scheduling options
    @click.option(
        "--type",
        metavar="SESSTYPE",
        type=click.Choice(["batch", "interactive"]),
        default="interactive",
        help="Either batch or interactive",
    )
    @click.option(
        "--starts-at",
        metavar="STARTS_AT",
        type=str,
        default=None,
        help="Let session to be started at a specific or relative time.",
    )
    @click.option(
        "-c",
        "--startup-command",
        metavar="COMMAND",
        help="Set the command to execute for batch-type sessions.",
    )
    @click.option(
        "--enqueue-only",
        is_flag=True,
        help="Enqueue the session and return immediately without waiting for its startup.",
    )
    @click.option(
        "--max-wait",
        metavar="SECONDS",
        type=int,
        default=0,
        help="The maximum duration to wait until the session starts.",
    )
    @click.option(
        "--no-reuse", is_flag=True, help="Do not reuse existing sessions but return an error."
    )
    @click.option(
        "--depends",
        metavar="SESSION_ID",
        type=str,
        multiple=True,
        help="Set the list of session ID or names that the newly created session depends on. "
        "The session will get scheduled after all of them successfully finish.",
    )
    @click.option(
        "--callback-url",
        metavar="CALLBACK_URL",
        type=str,
        default=None,
        help="Callback URL which will be called upon sesison lifecycle events.",
    )
    # execution environment
    @click.option(
        "-e",
        "--env",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Environment variable (may appear multiple times)",
    )
    # extra options
    @click.option(
        "--bootstrap-script",
        metavar="PATH",
        type=click.File("r"),
        default=None,
        help="A user-defined script to execute on startup.",
    )
    @click.option(
        "--tag", type=str, default=None, help="User-defined tag string to annotate sessions."
    )
    # resource spec
    @click.option(
        "-v",
        "--volume",
        "-m",
        "--mount",
        "mount",
        metavar="NAME[=PATH]",
        type=str,
        multiple=True,
        help="User-owned virtual folder names to mount. "
        "If path is not provided, virtual folder will be mounted under /home/work. "
        "When the target path is relative, it is placed under /home/work "
        "with auto-created parent directories if any. "
        "Absolute paths are mounted as-is, but it is prohibited to "
        "override the predefined Linux system directories.",
    )
    @click.option(
        "--scaling-group",
        "--sgroup",
        type=str,
        default=None,
        help="The scaling group to execute session. If not specified, "
        "all available scaling groups are included in the scheduling.",
    )
    @click.option(
        "-r",
        "--resources",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Set computation resources used by the session "
        "(e.g: -r cpu=2 -r mem=256 -r gpu=1)."
        "1 slot of cpu/gpu represents 1 core. "
        "The unit of mem(ory) is MiB.",
    )
    @click.option(
        "--cluster-size",
        metavar="NUMBER",
        type=int,
        default=1,
        help="The size of cluster in number of containers.",
    )
    @click.option(
        "--cluster-mode",
        metavar="MODE",
        type=click.Choice(["single-node", "multi-node"]),
        default="single-node",
        help="The mode of clustering.",
    )
    @click.option(
        "--resource-opts",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Resource options for creating compute session " "(e.g: shmem=64m)",
    )
    @click.option("--preopen", default=None, type=list_expr, help="Pre-open service ports")
    # resource grouping
    @click.option(
        "-d",
        "--domain",
        metavar="DOMAIN_NAME",
        default=None,
        help="Domain name where the session will be spawned. "
        "If not specified, config's domain name will be used.",
    )
    @click.option(
        "-g",
        "--group",
        metavar="GROUP_NAME",
        default=None,
        help="Group name where the session is spawned. "
        "User should be a member of the group to execute the code.",
    )
    @click.option(
        "--assign-agent",
        default=None,
        type=list_expr,
        help="Show mapping list of tuple which mapped containers with agent. "
        "When user role is Super Admin. "
        "(e.g., --assign-agent agent_id_1,agent_id_2,...)",
    )
    def create(
        # base args
        image: str,
        name: str | None,
        owner: str | None,
        # job scheduling options
        type: Literal["batch", "interactive"],
        starts_at: str | None,
        startup_command: str | None,
        enqueue_only: bool,
        max_wait: int,
        no_reuse: bool,
        depends: Sequence[str],
        callback_url: str,
        # execution environment
        env: Sequence[str],
        # extra options
        bootstrap_script: IO | None,
        tag: str | None,
        # resource spec
        mount: Sequence[str],
        scaling_group: str | None,
        resources: Sequence[str],
        cluster_size: int,
        cluster_mode: Literal["single-node", "multi-node"],
        resource_opts: Sequence[str],
        preopen: str | None,
        assign_agent: str | None,
        # resource grouping
        domain: str | None,
        group: str | None,
    ) -> None:
        """
        Prepare and start a single compute session without executing codes.
        You may use the created session to execute codes using the "run" command
        or connect to an application service provided by the session using the "app"
        command.


        \b
        IMAGE: The name (and version/platform tags appended after a colon) of session
               runtime or programming language.
        """
        if name is None:
            name = f"pysdk-{secrets.token_hex(5)}"
        else:
            name = name

        ######
        envs = prepare_env_arg(env)
        resources = prepare_resource_arg(resources)
        resource_opts = prepare_resource_arg(resource_opts)
        mount, mount_map = prepare_mount_arg(mount)

        preopen_ports = preopen
        assigned_agent_list = assign_agent
        with Session() as session:
            try:
                compute_session = session.ComputeSession.get_or_create(
                    image,
                    name=name,
                    type_=type,
                    starts_at=starts_at,
                    enqueue_only=enqueue_only,
                    max_wait=max_wait,
                    no_reuse=no_reuse,
                    dependencies=depends,
                    callback_url=callback_url,
                    cluster_size=cluster_size,
                    cluster_mode=cluster_mode,
                    mounts=mount,
                    mount_map=mount_map,
                    envs=envs,
                    startup_command=startup_command,
                    resources=resources,
                    resource_opts=resource_opts,
                    owner_access_key=owner,
                    domain_name=domain,
                    group_name=group,
                    scaling_group=scaling_group,
                    bootstrap_script=bootstrap_script.read()
                    if bootstrap_script is not None
                    else None,
                    tag=tag,
                    preopen_ports=preopen_ports,
                    assign_agent=assigned_agent_list,
                )
            except Exception as e:
                print_error(e)
                sys.exit(ExitCode.FAILURE)
            else:
                if compute_session.status == "PENDING":
                    print_info(
                        "Session ID {0} is enqueued for scheduling.".format(compute_session.id)
                    )
                elif compute_session.status == "SCHEDULED":
                    print_info(
                        "Session ID {0} is scheduled and about to be started.".format(
                            compute_session.id
                        )
                    )
                    return
                elif compute_session.status == "RUNNING":
                    if compute_session.created:
                        print_info(
                            "Session ID {0} is created and ready.".format(compute_session.id)
                        )
                    else:
                        print_info(
                            "Session ID {0} is already running and ready.".format(
                                compute_session.id
                            )
                        )
                    if compute_session.service_ports:
                        print_info(
                            "This session provides the following app services: "
                            + ", ".join(sport["name"] for sport in compute_session.service_ports)
                        )
                elif compute_session.status == "TERMINATED":
                    print_warn(
                        "Session ID {0} is already terminated.\n"
                        "This may be an error in the compute_session image.".format(
                            compute_session.id
                        )
                    )
                elif compute_session.status == "TIMEOUT":
                    print_info(
                        "Session ID {0} is still on the job queue.".format(compute_session.id)
                    )
                elif compute_session.status in ("ERROR", "CANCELLED"):
                    print_fail(
                        "Session ID {0} has an error during scheduling/startup or cancelled.".format(
                            compute_session.id
                        )
                    )

    if docs is not None:
        create.__doc__ = docs
    return create


main.command(aliases=["start"])(_create_cmd(docs='Alias of "session create"'))
session.command()(_create_cmd())


def _create_from_template_cmd(docs: str = None):
    @click.argument("template_id")
    @click.option(
        "-t",
        "--name",
        "--client-token",
        metavar="NAME",
        default=undefined,
        help="Specify a human-readable session name. " "If not set, a random hex string is used.",
    )
    @click.option(
        "-o",
        "--owner",
        "--owner-access-key",
        metavar="ACCESS_KEY",
        default=undefined,
        help="Set the owner of the target session explicitly.",
    )
    # job scheduling options
    @click.option(
        "--type",
        "type_",
        metavar="SESSTYPE",
        type=click.Choice(["batch", "interactive", undefined]),  # type: ignore
        default=undefined,
        help="Either batch or interactive",
    )
    @click.option(
        "--starts_at",
        metavar="STARTS_AT",
        type=str,
        default=None,
        help="Let session to be started at a specific or relative time.",
    )
    @click.option("-i", "--image", default=undefined, help="Set compute_session image to run.")
    @click.option(
        "-c",
        "--startup-command",
        metavar="COMMAND",
        default=undefined,
        help="Set the command to execute for batch-type sessions.",
    )
    @click.option(
        "--enqueue-only",
        is_flag=True,
        help="Enqueue the session and return immediately without waiting for its startup.",
    )
    @click.option(
        "--max-wait",
        metavar="SECONDS",
        type=int,
        default=undefined,
        help="The maximum duration to wait until the session starts.",
    )
    @click.option(
        "--no-reuse", is_flag=True, help="Do not reuse existing sessions but return an error."
    )
    @click.option(
        "--depends",
        metavar="SESSION_ID",
        type=str,
        multiple=True,
        help="Set the list of session ID or names that the newly created session depends on. "
        "The session will get scheduled after all of them successfully finish.",
    )
    @click.option(
        "--callback-url",
        metavar="CALLBACK_URL",
        type=str,
        default=None,
        help="Callback URL which will be called upon sesison lifecycle events.",
    )
    # execution environment
    @click.option(
        "-e",
        "--env",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Environment variable (may appear multiple times)",
    )
    # extra options
    @click.option(
        "--tag", type=str, default=undefined, help="User-defined tag string to annotate sessions."
    )
    # resource spec
    @click.option(
        "-m",
        "--mount",
        metavar="NAME[=PATH]",
        type=str,
        multiple=True,
        help="User-owned virtual folder names to mount. "
        "When the target path is relative, it is placed under /home/work "
        "with auto-created parent directories if any. "
        "Absolute paths are mounted as-is, but it is prohibited to "
        "override the predefined Linux system directories.",
    )
    @click.option(
        "--scaling-group",
        "--sgroup",
        type=str,
        default=undefined,
        help="The scaling group to execute session. If not specified, "
        "all available scaling groups are included in the scheduling.",
    )
    @click.option(
        "-r",
        "--resources",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Set computation resources used by the session "
        "(e.g: -r cpu=2 -r mem=256 -r gpu=1)."
        "1 slot of cpu/gpu represents 1 core. "
        "The unit of mem(ory) is MiB.",
    )
    @click.option(
        "--cluster-size",
        metavar="NUMBER",
        type=int,
        default=undefined,
        help="The size of cluster in number of containers.",
    )
    @click.option(
        "--resource-opts",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Resource options for creating compute session " "(e.g: shmem=64m)",
    )
    # resource grouping
    @click.option(
        "-d",
        "--domain",
        metavar="DOMAIN_NAME",
        default=None,
        help="Domain name where the session will be spawned. "
        "If not specified, config's domain name will be used.",
    )
    @click.option(
        "-g",
        "--group",
        metavar="GROUP_NAME",
        default=None,
        help="Group name where the session is spawned. "
        "User should be a member of the group to execute the code.",
    )
    # template overrides
    @click.option(
        "--no-mount",
        is_flag=True,
        help="If specified, client.py will tell server not to mount "
        "any vFolders specified at template,",
    )
    @click.option(
        "--no-env",
        is_flag=True,
        help="If specified, client.py will tell server not to add "
        "any environs specified at template,",
    )
    @click.option(
        "--no-resource",
        is_flag=True,
        help="If specified, client.py will tell server not to add "
        "any resource specified at template,",
    )
    def create_from_template(
        # base args
        template_id: str,
        name: str | Undefined,
        owner: str | Undefined,
        # job scheduling options
        type_: Literal["batch", "interactive"] | Undefined,
        starts_at: str | None,
        image: str | Undefined,
        startup_command: str | Undefined,
        enqueue_only: bool,
        max_wait: int | Undefined,
        no_reuse: bool,
        depends: Sequence[str],
        callback_url: str,
        # execution environment
        env: Sequence[str],
        # extra options
        tag: str | Undefined,
        # resource spec
        mount: Sequence[str],
        scaling_group: str | Undefined,
        resources: Sequence[str],
        cluster_size: int | Undefined,
        resource_opts: Sequence[str],
        # resource grouping
        domain: str | None,
        group: str | None,
        # template overrides
        no_mount: bool,
        no_env: bool,
        no_resource: bool,
    ) -> None:
        """
        Prepare and start a single compute session without executing codes.
        You may use the created session to execute codes using the "run" command
        or connect to an application service provided by the session using the "app"
        command.

        \b
        IMAGE: The name (and version/platform tags appended after a colon) of session
               runtime or programming language.
        """
        if name is undefined:
            name = f"pysdk-{secrets.token_hex(5)}"
        else:
            name = name

        envs = prepare_env_arg(env) if len(env) > 0 or no_env else undefined
        resources = (
            prepare_resource_arg(resources) if len(resources) > 0 or no_resource else undefined
        )
        resource_opts = (
            prepare_resource_arg(resource_opts)
            if len(resource_opts) > 0 or no_resource
            else undefined
        )
        prepared_mount, prepared_mount_map = (
            prepare_mount_arg(mount) if len(mount) > 0 or no_mount else (undefined, undefined)
        )
        with Session() as session:
            try:
                compute_session = session.ComputeSession.create_from_template(
                    template_id,
                    image=image,
                    name=name,
                    type_=type_,
                    starts_at=starts_at,
                    enqueue_only=enqueue_only,
                    max_wait=max_wait,
                    no_reuse=no_reuse,
                    dependencies=depends,
                    callback_url=callback_url,
                    cluster_size=cluster_size,
                    mounts=prepared_mount,
                    mount_map=prepared_mount_map,
                    envs=envs,
                    startup_command=startup_command,
                    resources=resources,
                    resource_opts=resource_opts,
                    owner_access_key=owner,
                    domain_name=domain,
                    group_name=group,
                    scaling_group=scaling_group,
                    tag=tag,
                )
            except Exception as e:
                print_error(e)
                sys.exit(ExitCode.FAILURE)
            else:
                if compute_session.status == "PENDING":
                    print_info("Session ID {0} is enqueued for scheduling.".format(name))
                elif compute_session.status == "SCHEDULED":
                    print_info("Session ID {0} is scheduled and about to be started.".format(name))
                    return
                elif compute_session.status == "RUNNING":
                    if compute_session.created:
                        print_info("Session ID {0} is created and ready.".format(name))
                    else:
                        print_info("Session ID {0} is already running and ready.".format(name))
                    if compute_session.service_ports:
                        print_info(
                            "This session provides the following app services: "
                            + ", ".join(sport["name"] for sport in compute_session.service_ports)
                        )
                elif compute_session.status == "TERMINATED":
                    print_warn(
                        "Session ID {0} is already terminated.\n"
                        "This may be an error in the compute_session image.".format(name)
                    )
                elif compute_session.status == "TIMEOUT":
                    print_info("Session ID {0} is still on the job queue.".format(name))
                elif compute_session.status in ("ERROR", "CANCELLED"):
                    print_fail(
                        "Session ID {0} has an error during scheduling/startup or cancelled.".format(
                            name
                        )
                    )

    if docs is not None:
        create_from_template.__doc__ = docs
    return create_from_template


main.command(aliases=["start-from-template"])(
    _create_from_template_cmd(docs='Alias of "session create-from-template"'),
)
session.command()(_create_from_template_cmd())


def _destroy_cmd(docs: str = None):
    @click.argument("session_names", metavar="SESSID", nargs=-1)
    @click.option(
        "-f",
        "--forced",
        is_flag=True,
        help="Force-terminate the errored sessions (only allowed for admins)",
    )
    @click.option(
        "-o",
        "--owner",
        "--owner-access-key",
        metavar="ACCESS_KEY",
        help="Specify the owner of the target session explicitly.",
    )
    @click.option(
        "-s", "--stats", is_flag=True, help="Show resource usage statistics after termination"
    )
    def destroy(session_names, forced, owner, stats):
        """
        Terminate and destroy the given session.

        SESSID: session ID given/generated when creating the session.
        """
        if len(session_names) == 0:
            print_warn('Specify at least one session ID. Check usage with "-h" option.')
            sys.exit(ExitCode.INVALID_ARGUMENT)
        print_wait("Terminating the session(s)...")
        with Session() as session:
            has_failure = False
            for session_name in session_names:
                try:
                    compute_session = session.ComputeSession(session_name, owner)
                    ret = compute_session.destroy(forced=forced)
                except BackendAPIError as e:
                    print_error(e)
                    if e.status == 404:
                        print_info(
                            'If you are an admin, use "-o" / "--owner" option '
                            "to terminate other user's session."
                        )
                    has_failure = True
                except Exception as e:
                    print_error(e)
                    has_failure = True
            else:
                if not has_failure:
                    print_done("Done.")
                if stats:
                    stats = ret.get("stats", None) if ret else None
                    if stats:
                        print(format_stats(stats))
                    else:
                        print("Statistics is not available.")
            if has_failure:
                sys.exit(ExitCode.FAILURE)

    if docs is not None:
        destroy.__doc__ = docs
    return destroy


main.command(aliases=["rm", "kill"])(_destroy_cmd(docs='Alias of "session destroy"'))
session.command(aliases=["rm", "kill"])(_destroy_cmd())


def _restart_cmd(docs: str = None):
    @click.argument("session_refs", metavar="SESSION_REFS", nargs=-1)
    def restart(session_refs):
        """
        Restart the compute session.

        \b
        SESSION_REF: session ID or name
        """
        if len(session_refs) == 0:
            print_warn('Specify at least one session ID. Check usage with "-h" option.')
            sys.exit(ExitCode.INVALID_ARGUMENT)
        print_wait("Restarting the session(s)...")
        with Session() as session:
            has_failure = False
            for session_ref in session_refs:
                try:
                    compute_session = session.ComputeSession(session_ref)
                    compute_session.restart()
                except BackendAPIError as e:
                    print_error(e)
                    if e.status == 404:
                        print_info(
                            'If you are an admin, use "-o" / "--owner" option '
                            "to terminate other user's session."
                        )
                    has_failure = True
                except Exception as e:
                    print_error(e)
                    has_failure = True
            else:
                if not has_failure:
                    print_done("Done.")
            if has_failure:
                sys.exit(ExitCode.FAILURE)

    if docs is not None:
        restart.__doc__ = docs
    return restart


main.command()(_restart_cmd(docs='Alias of "session restart"'))
session.command()(_restart_cmd())


@session.command()
@click.argument("session_id", metavar="SESSID")
@click.argument("files", type=click.Path(exists=True), nargs=-1)
def upload(session_id, files):
    """
    Upload the files to a compute session's home directory.
    If the target directory is in a storage folder mount, the operation is
    effectively same to uploading files to the storage folder.
    It is recommended to use storage folder commands for large file transfers
    to utilize the storage proxy.

    For cluster sessions, the files are only uploaded to the main container.

    \b
    SESSID: Session ID or name.
    FILES: One or more paths to upload.
    """
    if len(files) < 1:
        print_warn("Please specify one or more file paths after session ID or name.")
        return
    with Session() as session:
        try:
            print_wait("Uploading files...")
            kernel = session.ComputeSession(session_id)
            kernel.upload(files, show_progress=True)
            print_done("Uploaded.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command()
@click.argument("session_id", metavar="SESSID")
@click.argument("files", nargs=-1)
@click.option("--dest", type=Path, default=".", help="Destination path to store downloaded file(s)")
def download(session_id, files, dest):
    """
    Download files from a compute session's home directory.
    If the source path is in a storage folder mount, the operation is
    effectively same to downloading files from the storage folder.
    It is recommended to use storage folder commands for large file transfers
    to utilize the storage proxy.

    For cluster sessions, the files are only downloaded from the main container.

    \b
    SESSID: Session ID or name.
    FILES: One or more paths inside compute session.
    """
    if len(files) < 1:
        print_warn("Please specify one or more file paths after session ID or name.")
        return
    with Session() as session:
        try:
            print_wait("Downloading file(s) from {}...".format(session_id))
            kernel = session.ComputeSession(session_id)
            kernel.download(files, dest, show_progress=True)
            print_done("Downloaded to {}.".format(dest.resolve()))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command()
@click.argument("session_id", metavar="SESSID")
@click.argument("path", metavar="PATH", nargs=1, default="/home/work")
def ls(session_id, path):
    """
    List files in a path of a running compute session.

    For cluster sessions, it lists the files of the main container.

    \b
    SESSID: Session ID or name.
    PATH: Path inside container.
    """
    with Session() as session:
        try:
            print_wait('Retrieving list of files in "{}"...'.format(path))
            kernel = session.ComputeSession(session_id)
            result = kernel.list_files(path)

            if "errors" in result and result["errors"]:
                print_fail(result["errors"])
                sys.exit(ExitCode.FAILURE)

            files = json.loads(result["files"])
            table = []
            headers = ["File name", "Size", "Modified", "Mode"]
            for file in files:
                mdt = datetime.fromtimestamp(file["mtime"])
                fsize = naturalsize(file["size"], binary=True)
                mtime = mdt.strftime("%b %d %Y %H:%M:%S")
                row = [file["filename"], fsize, mtime, file["mode"]]
                table.append(row)
            print_done("Retrived.")
            print(tabulate(table, headers=headers))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command()
@click.argument("session_id", metavar="SESSID")
def logs(session_id):
    """
    Shows the full console log of a compute session.

    \b
    SESSID: Session ID or its alias given when creating the session.
    """
    with Session() as session:
        try:
            print_wait("Retrieving live container logs...")
            kernel = session.ComputeSession(session_id)
            result = kernel.get_logs().get("result")
            logs = result.get("logs") if "logs" in result else ""
            print(logs)
            print_done("End of logs.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command("status-history")
@click.argument("session_id", metavar="SESSID")
def status_history(session_id):
    """
    Shows the status transition history of the compute session.

    \b
    SESSID: Session ID or its alias given when creating the session.
    """
    with Session() as session:
        print_wait("Retrieving status history...")
        kernel = session.ComputeSession(session_id)
        try:
            status_history = kernel.get_status_history().get("result")
            print_info(f"status_history: {status_history}")
            if (preparing := status_history.get("preparing")) is None:
                result = {
                    "result": {
                        "seconds": 0,
                        "microseconds": 0,
                    },
                }
            elif (terminated := status_history.get("terminated")) is None:
                alloc_time_until_now: timedelta = datetime.now(tzutc()) - isoparse(preparing)
                result = {
                    "result": {
                        "seconds": alloc_time_until_now.seconds,
                        "microseconds": alloc_time_until_now.microseconds,
                    },
                }
            else:
                alloc_time: timedelta = isoparse(terminated) - isoparse(preparing)
                result = {
                    "result": {
                        "seconds": alloc_time.seconds,
                        "microseconds": alloc_time.microseconds,
                    },
                }
            print_done(f"Actual Resource Allocation Time: {result}")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command()
@click.argument("session_id", metavar="SESSID")
@click.argument("new_id", metavar="NEWID")
def rename(session_id, new_id):
    """
    Renames session name of running session.

    \b
    SESSID: Session ID or its alias given when creating the session.
    NEWID: New Session ID to rename to.
    """

    with Session() as session:
        try:
            kernel = session.ComputeSession(session_id)
            kernel.rename(new_id)
            print_done(f"Session renamed to {new_id}.")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command()
@click.argument("session_id", metavar="SESSID")
def commit(session_id):
    """
    Commit a running session to tar file.

    \b
    SESSID: Session ID or its alias given when creating the session.
    """

    with Session() as session:
        try:
            kernel = session.ComputeSession(session_id)
            kernel.commit()
            print_info(f"Request to commit Session(name or id: {session_id})")
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


@session.command()
@click.argument("session_id", metavar="SESSID")
def abuse_history(session_id):
    """
    Get abusing reports of session's sibling kernels.

    \b
    SESSID: Session ID or its alias given when creating the session.
    """

    with Session() as session:
        try:
            session = session.ComputeSession(session_id)
            report = session.get_abusing_report()
            print(dict(report))
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)


def _ssh_cmd(docs: str = None):
    @click.argument("session_ref", type=str, metavar="SESSION_REF")
    @click.option(
        "-p", "--port", type=int, metavar="PORT", default=9922, help="the port number for localhost"
    )
    @click.pass_context
    def ssh(ctx: click.Context, session_ref: str, port: int) -> None:
        """Execute the ssh command against the target compute session.

        \b
        SESSION_REF: The user-provided name or the unique ID of a running compute session.

        All remaining options and arguments not listed here are passed to the ssh command as-is.
        """
        try:
            with container_ssh_ctx(session_ref, port) as key_path:
                ssh_proc = subprocess.run(
                    [
                        "ssh",
                        "-o",
                        "StrictHostKeyChecking=no",
                        "-o",
                        "UserKnownHostsFile=/dev/null",
                        "-o",
                        "NoHostAuthenticationForLocalhost=yes",
                        "-i",
                        key_path,
                        "work@localhost",
                        "-p",
                        str(port),
                        *ctx.args,
                    ],
                    shell=False,
                    check=False,  # be transparent against the main command
                )
                sys.exit(ssh_proc.returncode)
        except Exception as e:
            print_error(e)

    if docs is not None:
        ssh.__doc__ = docs
    return ssh


_ssh_cmd_context_settings = {
    "ignore_unknown_options": True,
    "allow_extra_args": True,
    "allow_interspersed_args": True,
}

# Make it available as:
# - backend.ai ssh
# - backend.ai session ssh
main.command(
    context_settings=_ssh_cmd_context_settings,
)(_ssh_cmd(docs='Alias of "session ssh"'))
session.command(
    context_settings=_ssh_cmd_context_settings,
)(_ssh_cmd())


def _scp_cmd(docs: str = None):
    @click.argument("session_ref", type=str, metavar="SESSION_REF")
    @click.argument("src", type=str, metavar="SRC")
    @click.argument("dst", type=str, metavar="DST")
    @click.option(
        "-p", "--port", type=str, metavar="PORT", default=9922, help="the port number for localhost"
    )
    @click.option(
        "-r",
        "--recursive",
        default=False,
        is_flag=True,
        help="recursive flag option to process directories",
    )
    @click.pass_context
    def scp(
        ctx: click.Context,
        session_ref: str,
        src: str,
        dst: str,
        port: int,
        recursive: bool,
    ) -> None:
        """
        Execute the scp command against the target compute session.

        \b
        The SRC and DST have the same format with the original scp command,
        either a remote path as "work@localhost:path" or a local path.

        SESSION_REF: The user-provided name or the unique ID of a running compute session.
        SRC: the source path
        DST: the destination path

        All remaining options and arguments not listed here are passed to the ssh command as-is.

        Examples:

        * Uploading a local directory to the session:

          > backend.ai scp mysess -p 9922 -r tmp/ work@localhost:tmp2/

        * Downloading a directory from the session:

          > backend.ai scp mysess -p 9922 -r work@localhost:tmp2/ tmp/
        """
        recursive_args = []
        if recursive:
            recursive_args.append("-r")
        try:
            with container_ssh_ctx(session_ref, port) as key_path:
                scp_proc = subprocess.run(
                    [
                        "scp",
                        "-o",
                        "StrictHostKeyChecking=no",
                        "-o",
                        "UserKnownHostsFile=/dev/null",
                        "-o",
                        "NoHostAuthenticationForLocalhost=yes",
                        "-i",
                        key_path,
                        "-P",
                        str(port),
                        *recursive_args,
                        src,
                        dst,
                        *ctx.args,
                    ],
                    shell=False,
                    check=False,  # be transparent against the main command
                )
                sys.exit(scp_proc.returncode)
        except Exception as e:
            print_error(e)

    if docs is not None:
        scp.__doc__ = docs
    return scp


# Make it available as:
# - backend.ai scp
# - backend.ai session scp
main.command(
    context_settings=_ssh_cmd_context_settings,
)(_scp_cmd(docs='Alias of "session scp"'))
session.command(
    context_settings=_ssh_cmd_context_settings,
)(_scp_cmd())


def _events_cmd(docs: str = None):
    @click.argument("session_name_or_id", metavar="SESSION_ID_OR_NAME")
    @click.option(
        "-o",
        "--owner",
        "--owner-access-key",
        "owner_access_key",
        metavar="ACCESS_KEY",
        help="Specify the owner of the target session explicitly.",
    )
    @click.option(
        "--scope",
        type=click.Choice(["*", "session", "kernel"]),
        default="*",
        help="Filter the events by kernel-specific ones or session-specific ones.",
    )
    def events(session_name_or_id, owner_access_key, scope):
        """
        Monitor the lifecycle events of a compute session.

        SESSID: session ID or its alias given when creating the session.
        """

        async def _run_events():
            async with AsyncSession() as session:
                try:
                    session_id = uuid.UUID(session_name_or_id)
                    compute_session = session.ComputeSession.from_session_id(session_id)
                except ValueError:
                    compute_session = session.ComputeSession(session_name_or_id, owner_access_key)
                async with compute_session.listen_events(scope=scope) as response:
                    async for ev in response:
                        click.echo(
                            click.style(ev.event, fg="cyan", bold=True)
                            + " "
                            + json.dumps(json.loads(ev.data), indent=None)  # as single-line
                        )

        try:
            asyncio_run(_run_events())
        except Exception as e:
            print_error(e)

    if docs is not None:
        events.__doc__ = docs
    return events


# Make it available as:
# - backend.ai events
# - backend.ai session events
main.command()(_events_cmd(docs='Alias of "session events"'))
session.command()(_events_cmd())


def _fetch_session_names():
    status = ",".join(
        [
            "PENDING",
            "SCHEDULED",
            "PREPARING",
            "PULLING",
            "RUNNING",
            "RESTARTING",
            "TERMINATING",
            "RESIZING",
            "SUSPENDED",
            "ERROR",
        ]
    )
    fields: List[FieldSpec] = [
        session_fields["name"],
        session_fields["session_id"],
        session_fields["group_name"],
        session_fields["kernel_id"],
        session_fields["image"],
        session_fields["type"],
        session_fields["status"],
        session_fields["status_info"],
        session_fields["status_changed"],
        session_fields["result"],
    ]
    with Session() as session:
        sessions = session.ComputeSession.paginated_list(
            status=status,
            access_key=None,
            fields=fields,
            page_offset=0,
            page_size=10,
            filter=None,
            order=None,
        )

    return tuple(map(lambda x: x.get("session_id"), sessions.items))


def _watch_cmd(docs: Optional[str] = None):
    @click.argument("session_name_or_id", metavar="SESSION_ID_OR_NAME", nargs=-1)
    @click.option(
        "-o",
        "--owner",
        "--owner-access-key",
        "owner_access_key",
        metavar="ACCESS_KEY",
        help="Specify the owner of the target session explicitly.",
    )
    @click.option(
        "--scope",
        type=click.Choice(["*", "session", "kernel"]),
        default="*",
        help="Filter the events by kernel-specific ones or session-specific ones.",
    )
    @click.option(
        "--max-wait",
        metavar="SECONDS",
        type=int,
        default=0,
        help="The maximum duration to wait until the session starts.",
    )
    @click.option(
        "--output",
        type=click.Choice(["json", "console"]),
        default="console",
        help="Set the output style of the command results.",
    )
    def watch(
        session_name_or_id: str, owner_access_key: str, scope: str, max_wait: int, output: str
    ):
        """
        Monitor the lifecycle events of a compute session
        and display in human-friendly interface.
        """
        session_names = _fetch_session_names()
        if not session_names:
            if output == "json":
                sys.stderr.write(f'{json.dumps({"ok": False, "reason": "No matching items."})}\n')
            else:
                print_fail("No matching items.")
            sys.exit(ExitCode.FAILURE)

        if not session_name_or_id:
            questions = [
                inquirer.List(
                    "session",
                    message="Select session to watch.",
                    choices=session_names,
                )
            ]
            session_name_or_id = inquirer.prompt(questions).get("session")
        else:
            for session_name in session_names:
                if session_name.startswith(session_name_or_id[0]):
                    session_name_or_id = session_name
                    break
            else:
                if output == "json":
                    sys.stderr.write(
                        f'{json.dumps({"ok": False, "reason": "No matching items."})}\n'
                    )
                else:
                    print_fail("No matching items.")
                sys.exit(ExitCode.FAILURE)

        async def handle_console_output(
            session: ComputeSession, scope: Literal["*", "session", "kernel"] = "*"
        ):
            async with session.listen_events(scope=scope) as response:  # AsyncSession
                async for ev in response:
                    match ev.event:
                        case events.SESSION_SUCCESS:
                            print_done(events.SESSION_SUCCESS)
                            sys.exit(json.loads(ev.data).get("exitCode", 0))
                        case events.SESSION_FAILURE:
                            print_fail(events.SESSION_FAILURE)
                            sys.exit(json.loads(ev.data).get("exitCode", 1))
                        case events.KERNEL_CANCELLED:
                            print_fail(events.KERNEL_CANCELLED)
                            break
                        case events.SESSION_TERMINATED:
                            print_done(events.SESSION_TERMINATED)
                            break
                        case _:
                            print_done(ev.event)

        async def handle_json_output(
            session: ComputeSession, scope: Literal["*", "session", "kernel"] = "*"
        ):
            async with session.listen_events(scope=scope) as response:  # AsyncSession
                async for ev in response:
                    event = json.loads(ev.data)
                    event["event"] = ev.event
                    click.echo(event)

                    match ev.event:
                        case events.SESSION_SUCCESS:
                            sys.exit(event.get("exitCode", 0))
                        case events.SESSION_FAILURE:
                            sys.exit(event.get("exitCode", 1))
                        case events.SESSION_TERMINATED | events.KERNEL_CANCELLED:
                            break

        async def _run_events():
            async with AsyncSession() as session:
                try:
                    session_id = uuid.UUID(session_name_or_id)
                    compute_session = session.ComputeSession.from_session_id(session_id)
                except ValueError:
                    compute_session = session.ComputeSession(session_name_or_id, owner_access_key)

                if output == "console":
                    await handle_console_output(session=compute_session, scope=scope)
                elif output == "json":
                    await handle_json_output(session=compute_session, scope=scope)

        async def _run_events_with_timeout(max_wait: int):
            try:
                async with timeout(max_wait):
                    await _run_events()
            except asyncio.TimeoutError:
                sys.exit(ExitCode.OPERATION_TIMEOUT)

        try:
            if max_wait > 0:
                asyncio_run(_run_events_with_timeout(max_wait))
            else:
                asyncio_run(_run_events())
        except Exception as e:
            print_error(e)
            sys.exit(ExitCode.FAILURE)

    if docs is not None:
        watch.__doc__ = docs
    return watch


# Make it available as:
# - backend.ai watch
# - backend.ai session watch
main.command()(_watch_cmd(docs='Alias of "session watch"'))
session.command()(_watch_cmd())



import asyncio
from pathlib import Path
from typing import Dict, Mapping, Optional, Sequence, Union

import aiohttp
import janus
from aiohttp import hdrs
from aiotusclient import client
from tenacity import (
    AsyncRetrying,
    RetryError,
    TryAgain,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from tqdm import tqdm
from yarl import URL

from ai.backend.client.output.fields import vfolder_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult

from ..compat import current_loop
from ..config import DEFAULT_CHUNK_SIZE, MAX_INFLIGHT_CHUNKS
from ..exceptions import BackendClientError
from ..pagination import fetch_paginated_result
from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("VFolder",)

_default_list_fields = (
    vfolder_fields["host"],
    vfolder_fields["name"],
    vfolder_fields["created_at"],
    vfolder_fields["creator"],
    vfolder_fields["group_id"],
    vfolder_fields["permission"],
    vfolder_fields["ownership_type"],
    vfolder_fields["status"],
)


class ResponseFailed(Exception):
    pass


class VFolder(BaseFunction):
    def __init__(self, name: str):
        self.name = name

    @api_function
    @classmethod
    async def create(
        cls,
        name: str,
        host: str = None,
        unmanaged_path: str = None,
        group: str = None,
        usage_mode: str = "general",
        permission: str = "rw",
        quota: str = "0",
        cloneable: bool = False,
    ):
        rqst = Request("POST", "/folders")
        rqst.set_json(
            {
                "name": name,
                "host": host,
                "unmanaged_path": unmanaged_path,
                "group": group,
                "usage_mode": usage_mode,
                "permission": permission,
                "quota": quota,
                "cloneable": cloneable,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def delete_by_id(cls, oid):
        rqst = Request("DELETE", "/folders")
        rqst.set_json({"id": oid})
        async with rqst.fetch():
            return {}

    @api_function
    @classmethod
    async def list(cls, list_all=False):
        rqst = Request("GET", "/folders")
        rqst.set_json({"all": list_all})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        group: str = None,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of vfolders. Domain admins can only get domain vfolders.

        :param group: Fetch vfolders in a specific group.
        :param fields: Additional per-vfolder query fields to fetch.
        """
        return await fetch_paginated_result(
            "vfolder_list",
            {
                "group_id": (group, "UUID"),
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def list_hosts(cls):
        rqst = Request("GET", "/folders/_/hosts")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_all_hosts(cls):
        rqst = Request("GET", "/folders/_/all_hosts")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_allowed_types(cls):
        rqst = Request("GET", "/folders/_/allowed_types")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def info(self):
        rqst = Request("GET", "/folders/{0}".format(self.name))
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete(self):
        rqst = Request("DELETE", "/folders/{0}".format(self.name))
        async with rqst.fetch():
            return {}

    @api_function
    async def rename(self, new_name):
        rqst = Request("POST", "/folders/{0}/rename".format(self.name))
        rqst.set_json(
            {
                "new_name": new_name,
            }
        )
        async with rqst.fetch() as resp:
            self.name = new_name
            return await resp.text()

    def _write_file(self, file_path: Path, mode: str, q: janus._SyncQueueProxy[bytes]):
        with open(file_path, mode) as f:
            while True:
                chunk = q.get()
                if not chunk:
                    return
                f.write(chunk)
                q.task_done()

    async def _download_file(
        self,
        file_path: Path,
        download_url: URL,
        chunk_size: int,
        max_retries: int,
        show_progress: bool,
    ) -> None:
        if show_progress:
            print(f"Downloading to {file_path} ...")

        range_start = 0
        if_range: str | None = None
        file_unit = "bytes"
        file_mode = "wb"
        file_req_hdrs: Dict[str, str] = {}
        try:
            async for session_attempt in AsyncRetrying(
                wait=wait_exponential(multiplier=0.02, min=0.02, max=5.0),
                stop=stop_after_attempt(max_retries),
                retry=retry_if_exception_type(TryAgain),
            ):
                with session_attempt:
                    try:
                        if if_range is not None:
                            file_req_hdrs[hdrs.IF_RANGE] = if_range
                            file_req_hdrs[hdrs.RANGE] = f"{file_unit}={range_start}-"
                        async with aiohttp.ClientSession(headers=file_req_hdrs) as client:
                            async with client.get(download_url, ssl=False) as raw_resp:
                                match raw_resp.status:
                                    case 200:
                                        # First attempt to download file or file has changed.
                                        file_mode = "wb"
                                        range_start = 0
                                    case 206:
                                        # File has not changed. Continue downloading from range_start.
                                        file_mode = "ab"
                                    case _:
                                        # Retry.
                                        raise ResponseFailed
                                size = int(raw_resp.headers["Content-Length"])
                                if_range = raw_resp.headers.get("Last-Modified")
                                q: janus.Queue[bytes] = janus.Queue(MAX_INFLIGHT_CHUNKS)
                                try:
                                    with tqdm(
                                        total=(size - range_start),
                                        unit=file_unit,
                                        unit_scale=True,
                                        unit_divisor=1024,
                                        disable=not show_progress,
                                    ) as pbar:
                                        loop = current_loop()
                                        writer_fut = loop.run_in_executor(
                                            None, self._write_file, file_path, file_mode, q.sync_q
                                        )
                                        await asyncio.sleep(0)
                                        max_attempts = 10
                                        while True:
                                            try:
                                                async for attempt in AsyncRetrying(
                                                    wait=wait_exponential(
                                                        multiplier=0.02, min=0.02, max=5.0
                                                    ),
                                                    stop=stop_after_attempt(max_attempts),
                                                    retry=retry_if_exception_type(TryAgain),
                                                ):
                                                    with attempt:
                                                        try:
                                                            chunk = await raw_resp.content.read(
                                                                chunk_size
                                                            )
                                                        except asyncio.TimeoutError:
                                                            raise TryAgain
                                            except RetryError:
                                                raise ResponseFailed
                                            range_start += len(chunk)
                                            pbar.update(len(chunk))
                                            if not chunk:
                                                break
                                            await q.async_q.put(chunk)
                                finally:
                                    await q.async_q.put(b"")
                                    await writer_fut
                                    q.close()
                                    await q.wait_closed()
                    except (
                        ResponseFailed,
                        aiohttp.ClientPayloadError,
                        aiohttp.ClientConnectorError,
                    ):
                        raise TryAgain
        except RetryError:
            raise RuntimeError(f"Downloading {file_path.name} failed after {max_retries} retries")

    @api_function
    async def download(
        self,
        relative_paths: Sequence[Union[str, Path]],
        *,
        basedir: Union[str, Path] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        show_progress: bool = False,
        address_map: Optional[Mapping[str, str]] = None,
        max_retries: int = 20,
    ) -> None:
        base_path = Path.cwd() if basedir is None else Path(basedir).resolve()
        for relpath in relative_paths:
            file_path = base_path / relpath
            if file_path.exists():
                raise RuntimeError("The target file already exists", file_path.name)
            rqst = Request("POST", "/folders/{}/request-download".format(self.name))
            rqst.set_json(
                {
                    "path": str(relpath),
                }
            )
            async with rqst.fetch() as resp:
                download_info = await resp.json()
                overriden_url = download_info["url"]
                if address_map:
                    if download_info["url"] in address_map:
                        overriden_url = address_map[download_info["url"]]
                    else:
                        raise BackendClientError(
                            "Overriding storage proxy addresses are given, "
                            "but no url matches with any of them.\n",
                        )

                download_url = URL(overriden_url).with_query(
                    {
                        "token": download_info["token"],
                    }
                )
            await self._download_file(
                file_path, download_url, chunk_size, max_retries, show_progress
            )

    @api_function
    async def upload(
        self,
        files: Sequence[Union[str, Path]],
        *,
        basedir: Union[str, Path] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        address_map: Optional[Mapping[str, str]] = None,
        show_progress: bool = False,
    ) -> None:
        base_path = Path.cwd() if basedir is None else Path(basedir).resolve()
        if basedir:
            files = [basedir / Path(file) for file in files]
        else:
            files = [Path(file).resolve() for file in files]
        for file_path in files:
            file_size = Path(file_path).stat().st_size
            rqst = Request("POST", "/folders/{}/request-upload".format(self.name))
            rqst.set_json(
                {
                    "path": "{}".format(str(Path(file_path).relative_to(base_path))),
                    "size": int(file_size),
                }
            )
            async with rqst.fetch() as resp:
                upload_info = await resp.json()
                overriden_url = upload_info["url"]
                if address_map:
                    if upload_info["url"] in address_map:
                        overriden_url = address_map[upload_info["url"]]
                    else:
                        raise BackendClientError(
                            "Overriding storage proxy addresses are given, "
                            "but no url matches with any of them.\n",
                        )
                upload_url = URL(overriden_url).with_query(
                    {
                        "token": upload_info["token"],
                    }
                )
            tus_client = client.TusClient()
            if basedir:
                input_file = open(base_path / file_path, "rb")
            else:
                input_file = open(str(Path(file_path).relative_to(base_path)), "rb")
            print(f"Uploading {base_path / file_path} via {upload_info['url']} ...")
            # TODO: refactor out the progress bar
            uploader = tus_client.async_uploader(
                file_stream=input_file,
                url=upload_url,
                upload_checksum=False,
                chunk_size=chunk_size,
            )
            return await uploader.upload()

    @api_function
    async def mkdir(
        self,
        path: Union[str, Path],
        parents: Optional[bool] = False,
        exist_ok: Optional[bool] = False,
    ):
        rqst = Request("POST", "/folders/{}/mkdir".format(self.name))
        rqst.set_json(
            {
                "path": path,
                "parents": parents,
                "exist_ok": exist_ok,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    async def rename_file(self, target_path: str, new_name: str):
        rqst = Request("POST", "/folders/{}/rename-file".format(self.name))
        rqst.set_json(
            {
                "target_path": target_path,
                "new_name": new_name,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def move_file(self, src_path: str, dst_path: str):
        rqst = Request("POST", "/folders/{}/move-file".format(self.name))
        rqst.set_json(
            {
                "src": src_path,
                "dst": dst_path,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete_files(self, files: Sequence[Union[str, Path]], recursive: bool = False):
        rqst = Request("DELETE", "/folders/{}/delete-files".format(self.name))
        rqst.set_json(
            {
                "files": files,
                "recursive": recursive,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    async def list_files(self, path: Union[str, Path] = "."):
        rqst = Request("GET", "/folders/{}/files".format(self.name))
        rqst.set_json(
            {
                "path": path,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def invite(self, perm: str, emails: Sequence[str]):
        rqst = Request("POST", "/folders/{}/invite".format(self.name))
        rqst.set_json(
            {
                "perm": perm,
                "user_ids": emails,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def invitations(cls):
        rqst = Request("GET", "/folders/invitations/list")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def accept_invitation(cls, inv_id: str):
        rqst = Request("POST", "/folders/invitations/accept")
        rqst.set_json({"inv_id": inv_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def delete_invitation(cls, inv_id: str):
        rqst = Request("DELETE", "/folders/invitations/delete")
        rqst.set_json({"inv_id": inv_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_fstab_contents(cls, agent_id=None):
        rqst = Request("GET", "/folders/_/fstab")
        rqst.set_json(
            {
                "agent_id": agent_id,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_performance_metric(cls, folder_host: str):
        rqst = Request("GET", "/folders/_/perf-metric")
        rqst.set_json(
            {
                "folder_host": folder_host,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_mounts(cls):
        rqst = Request("GET", "/folders/_/mounts")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def mount_host(cls, name: str, fs_location: str, options=None, edit_fstab: bool = False):
        rqst = Request("POST", "/folders/_/mounts")
        rqst.set_json(
            {
                "name": name,
                "fs_location": fs_location,
                "options": options,
                "edit_fstab": edit_fstab,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def umount_host(cls, name: str, edit_fstab: bool = False):
        rqst = Request("DELETE", "/folders/_/mounts")
        rqst.set_json(
            {
                "name": name,
                "edit_fstab": edit_fstab,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def share(self, perm: str, emails: Sequence[str]):
        rqst = Request("POST", "/folders/{}/share".format(self.name))
        rqst.set_json(
            {
                "permission": perm,
                "emails": emails,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def unshare(self, emails: Sequence[str]):
        rqst = Request("DELETE", "/folders/{}/unshare".format(self.name))
        rqst.set_json(
            {
                "emails": emails,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def leave(self, shared_user_uuid=None):
        rqst = Request("POST", "/folders/{}/leave".format(self.name))
        rqst.set_json({"shared_user_uuid": shared_user_uuid})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def clone(
        self,
        target_name: str,
        target_host: str = None,
        usage_mode: str = "general",
        permission: str = "rw",
    ):
        rqst = Request("POST", "/folders/{}/clone".format(self.name))
        rqst.set_json(
            {
                "target_name": target_name,
                "target_host": target_host,
                "usage_mode": usage_mode,
                "permission": permission,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def update_options(self, name: str, permission: str = None, cloneable: bool = None):
        rqst = Request("POST", "/folders/{}/update-options".format(self.name))
        rqst.set_json(
            {
                "cloneable": cloneable,
                "permission": permission,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    @classmethod
    async def list_shared_vfolders(cls):
        rqst = Request("GET", "folders/_/shared")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def shared_vfolder_info(cls, vfolder_id: str):
        rqst = Request("GET", "folders/_/shared")
        rqst.set_json({"vfolder_id": vfolder_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def update_shared_vfolder(cls, vfolder: str, user: str, perm: str = None):
        rqst = Request("POST", "/folders/_/shared")
        rqst.set_json(
            {
                "vfolder": vfolder,
                "user": user,
                "perm": perm,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()



from __future__ import annotations

import sys

import click

from ai.backend.cli.interaction import ask_yn
from ai.backend.cli.types import ExitCode
from ai.backend.client.output.fields import user_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..params import CommaSeparatedListType
from ..pretty import print_info
from ..types import CLIContext
from . import admin

list_expr = CommaSeparatedListType()


@admin.group()
def user() -> None:
    """
    User administration commands.
    """


@user.command()
@pass_ctx_obj
@click.option("-e", "--email", type=str, default=None, help="Email of a user to display.")
def info(ctx: CLIContext, email: str) -> None:
    """
    Show the information about the given user by email. If email is not give,
    requester's information will be displayed.
    """
    fields = [
        user_fields["uuid"],
        user_fields["username"],
        user_fields["role"],
        user_fields["email"],
        user_fields["full_name"],
        user_fields["need_password_change"],
        user_fields["status"],
        user_fields["status_info"],
        user_fields["created_at"],
        user_fields["domain_name"],
        user_fields["groups"],
        user_fields["allowed_client_ip"],
    ]
    with Session() as session:
        try:
            item = session.User.detail(email=email, fields=fields)
            ctx.output.print_item(item, fields=fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@user.command()
@pass_ctx_obj
@click.option(
    "-s",
    "--status",
    type=str,
    default=None,
    help="Filter users in a specific state (active, inactive, deleted, before-verification).",
)
@click.option("-g", "--group", type=str, default=None, help="Filter by group ID.")
@click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
@click.option("--order", default=None, help="Set the query ordering expression.")
@click.option("--offset", default=0, help="The index of the current page start for pagination.")
@click.option("--limit", type=int, default=None, help="The page size for pagination.")
def list(ctx: CLIContext, status, group, filter_, order, offset, limit) -> None:
    """
    List users.
    (admin privilege required)
    """
    fields = [
        user_fields["uuid"],
        user_fields["username"],
        user_fields["role"],
        user_fields["email"],
        user_fields["full_name"],
        user_fields["need_password_change"],
        user_fields["status"],
        user_fields["status_info"],
        user_fields["created_at"],
        user_fields["domain_name"],
        user_fields["groups"],
        user_fields["allowed_client_ip"],
    ]
    try:
        with Session() as session:
            fetch_func = lambda pg_offset, pg_size: session.User.paginated_list(
                status,
                group,
                fields=fields,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
    except Exception as e:
        ctx.output.print_error(e)
        sys.exit(ExitCode.FAILURE)


@user.command()
@pass_ctx_obj
@click.argument("domain_name", type=str, metavar="DOMAIN_NAME")
@click.argument("email", type=str, metavar="EMAIL")
@click.argument("password", type=str, metavar="PASSWORD")
@click.option("-u", "--username", type=str, default="", help="Username.")
@click.option("-n", "--full-name", type=str, default="", help="Full name.")
@click.option(
    "-r",
    "--role",
    type=str,
    default="user",
    help="Role of the user. One of (admin, user, monitor).",
)
@click.option(
    "-s",
    "--status",
    type=str,
    default="active",
    help="Account status. One of (active, inactive, deleted, before-verification).",
)
@click.option(
    "--need-password-change",
    is_flag=True,
    help="Flag indicate that user needs to change password. "
    "Useful when admin manually create password.",
)
@click.option(
    "--allowed-ip",
    type=list_expr,
    default=None,
    help="Allowed client IP. IPv4 and IPv6 are allowed. CIDR type is recommended. "
    '(e.g., --allowed-ip "127.0.0.1","127.0.0.2",...)',
)
@click.option("--description", type=str, default="", help="Description of the user.")
def add(
    ctx: CLIContext,
    domain_name,
    email,
    password,
    username,
    full_name,
    role,
    status,
    need_password_change,
    allowed_ip,
    description,
):
    """
    Add new user. A user must belong to a domain, so DOMAIN_NAME should be provided.

    \b
    DOMAIN_NAME: Name of the domain where new user belongs to.
    EMAIL: Email of new user.
    PASSWORD: Password of new user.
    """
    with Session() as session:
        try:
            data = session.User.create(
                domain_name,
                email,
                password,
                username=username,
                full_name=full_name,
                role=role,
                status=status,
                need_password_change=need_password_change,
                allowed_client_ip=allowed_ip,
                description=description,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            item_name="user",
        )


@user.command()
@pass_ctx_obj
@click.argument("email", type=str, metavar="EMAIL")
@click.option("-p", "--password", type=str, help="Password.")
@click.option("-u", "--username", type=str, help="Username.")
@click.option("-n", "--full-name", type=str, help="Full name.")
@click.option("-d", "--domain-name", type=str, help="Domain name.")
@click.option(
    "-r",
    "--role",
    type=str,
    help="Role of the user. One of (admin, user, monitor).",
)
@click.option(
    "-s",
    "--status",
    type=str,
    help="Account status. One of (active, inactive, deleted, before-verification).",
)
@click.option(
    "--need-password-change",
    is_flag=True,
    help="Flag indicate that user needs to change password. "
    "Useful when admin manually create password.",
)
@click.option(
    "--allowed-ip",
    type=list_expr,
    default=None,
    help="Allowed client IP. IPv4 and IPv6 are allowed. CIDR type is recommended. "
    '(e.g., --allowed-ip "127.0.0.1","127.0.0.2",...)',
)
@click.option("--description", type=str, default="", help="Description of the user.")
def update(
    ctx: CLIContext,
    email,
    password,
    username,
    full_name,
    domain_name,
    role,
    status,
    need_password_change,
    allowed_ip,
    description,
):
    """
    Update an existing user.

    EMAIL: Email of user to update.
    """
    with Session() as session:
        try:
            data = session.User.update(
                email,
                password=password,
                username=username,
                full_name=full_name,
                domain_name=domain_name,
                role=role,
                status=status,
                need_password_change=need_password_change,
                allowed_client_ip=allowed_ip,
                description=description,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "email": email,
            },
        )


@user.command()
@pass_ctx_obj
@click.argument("email", type=str, metavar="EMAIL")
def delete(ctx: CLIContext, email):
    """
    Inactivate an existing user.

    EMAIL: Email of user to inactivate.
    """
    with Session() as session:
        try:
            data = session.User.delete(email)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "email": email,
            },
        )


@user.command()
@pass_ctx_obj
@click.argument("email", type=str, metavar="EMAIL")
@click.option(
    "--purge-shared-vfolders",
    is_flag=True,
    default=False,
    help="Delete user's all virtual folders. "
    "If False, shared folders will not be deleted "
    "and migrated the ownership to the requested admin.",
)
def purge(ctx: CLIContext, email, purge_shared_vfolders):
    """
    Delete an existing user. This action cannot be undone.

    NAME: Name of a domain to delete.
    """
    with Session() as session:
        try:
            if not ask_yn():
                print_info("Cancelled")
                sys.exit(ExitCode.FAILURE)
            data = session.User.purge(email, purge_shared_vfolders)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="user",
                action_name="purge",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="user",
                action_name="purge",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "email": email,
            },
        )
